from __future__ import absolute_import
import py


def test_pypath(q):
    p = py.path.local('/abc/def')
    # NB: On Windows, the result is ":C:/abc/def"
    assert str(q('::', p)).endswith(':/abc/def')



from __future__ import absolute_import

import pytest
import sys
import os
import unittest
from datetime import datetime, date, time, timedelta
import math
import struct
from pyq import _k, Q_VERSION, Q_OS

_PY3K = 1

WIN = Q_OS.startswith('w')
if sys.maxsize == 2147483647 or WIN:  # 32-bit Unix or any Windows
    K_INT_CODE, K_LONG_CODE = "lq"
else:
    K_INT_CODE, K_LONG_CODE = "il"


# extract _k.K class methods
for m in ('func k knk ktd err dot a1'
          ' ka kb kg kh ki kj ke kf kc ks km kd kz ku kv kt kp'
          ' knt ktn ktj kzz knz kpz b9 d9'
          ' B G H I J E F S K P M D N U V T xT xD').split():
    globals()[m] = getattr(_k.K, '_' + m, None)
del m


def q(*args):
    return k(0, *args)


q("\\e 0")  # disable q's debug on error

if Q_VERSION >= 3:
    kguid = _k.K._kguid
    UU = _k.K._UU


def kstr(x):
    return k(0, '-3!', x).inspect(b's')


class K_TestCase(unittest.TestCase):
    def assert_k_is(self, x, y):
        test = k(0, '~[%s;]' % y, x).inspect(b'g')
        if not test:
            raise self.failureException(kstr(x) + ' <> ' + y)


class AtomTestCase(K_TestCase):
    def test_ka(self):
        self.assert_k_is(ka(-_k.KB, 1), '1b')
        self.assert_k_is(ka(-_k.KJ, 1), '1j')
        self.assert_k_is(ktj(-_k.KP, 0), '2000.01.01D00:00:00.000000000')
        self.assert_k_is(ktj(-_k.KN, 0), '0D00:00:00.000000000')

    def test_kb(self):
        self.assert_k_is(kb(1), '1b')
        self.assert_k_is(kb(2 ** 70), '1b')
        self.assert_k_is(kb(True), '1b')
        self.assert_k_is(kb(0), '0b')
        self.assert_k_is(kb(False), '0b')
        self.assert_k_is(kb(None), '0b')

    def test_kg(self):
        self.assert_k_is(kg(1), '0x01')

    def test_kh(self):
        self.assert_k_is(kh(1), '1h')

    def test_ki(self):
        self.assert_k_is(ki(1), '1i')

    def test_kj(self):
        self.assert_k_is(kj(1), '1j')
        self.assert_k_is(kj(9223372036854775806), '9223372036854775806j')

    def test_ke(self):
        self.assert_k_is(ke(1), '1e')

    def test_kf(self):
        self.assert_k_is(kf(1), '1f')

    def test_kc(self):
        self.assert_k_is(kc(b'x'), '"x"')
        self.assert_k_is(kc('x'), '"x"')

    def test_ks(self):
        self.assert_k_is(ks('abc'), '`abc')

    def test_km(self):
        self.assert_k_is(km(1), '2000.02m')

    def test_kd(self):
        self.assert_k_is(kd(1), '2000.01.02')

    def test_kzz(self):
        self.assert_k_is(kzz(datetime(2000, 1, 2, 10, 11, 12, 1000)),
                         '2000.01.02T10:11:12.001')

    def test_kpz(self):
        self.assert_k_is(kpz(datetime(2000, 1, 2, 10, 11, 12, 1000)),
                         '2000.01.02D10:11:12.001000000')

    def test_knz(self):
        self.assert_k_is(knz(timedelta(100, 20, 1000)),
                         '100D00:00:20.001000000')

    def test_kz(self):
        self.assert_k_is(kz(1.5), '2000.01.02T12:00:00.000')

    def test_ku(self):
        self.assert_k_is(ku(1), '00:01')

    def test_kv(self):
        self.assert_k_is(kv(1), '00:00:01')

    def test_kt(self):
        self.assert_k_is(kt(1), '00:00:00.001')

    def test_id(self):
        x, y = ki(1), ki(1)
        self.assertNotEqual(x._id(), y._id())


class ListTestCase(K_TestCase):
    def test_kp(self):
        self.assert_k_is(kp("abc"), '"abc"')

    def test_I(self):
        self.assert_k_is(I([]), '`int$()')
        self.assert_k_is(I([1, 2]), '1 2i')

    def test_J(self):
        self.assert_k_is(J([]), '`long$()')
        self.assert_k_is(J([1, 2]), '1 2j')

    def test_F(self):
        self.assert_k_is(F([]), '`float$()')
        self.assert_k_is(F([1., 2.]), '1 2f')

    def test_S(self):
        self.assert_k_is(S([]), '`symbol$()')
        self.assert_k_is(S(['aa', 'bb', u'uu', None]), '`aa`bb`uu`')

    def test_K(self):
        self.assert_k_is(K([]), '()')
        self.assert_k_is(K([ki(0), kf(1)]), '(0i;1f)')


class TableDictTestCase(K_TestCase):
    def test_xD(self):
        self.assert_k_is(xD(S('abc'), I(range(3))), '`a`b`c!0 1 2i')

    def test_xT(self):
        a = S('XYZ')
        b = kp('xyz')
        c = I([1, 2, 3])
        self.assert_k_is(xT(xD(S('abc'), K([a, b, c]))),
                         '([]a:`X`Y`Z;b:"xyz";c:1 2 3i)')

    def test_ktd(self):
        x = k(0, '([a:1 2 3]b:10 20 30)')
        y = '([]a:1 2 3;b:10 20 30)'
        self.assert_k_is(ktd(x), y)


class IterTestCase(K_TestCase):
    def test_bool(self):
        self.assertEqual(list(k(0, '101b')), [True, False, True])

    def test_byte(self):
        self.assertEqual(list(k(0, '0x010203')), [1, 2, 3])

    def test_char(self):
        self.assertEqual(list(kp('abc')), ['a', 'b', 'c'])

    def test_short(self):
        self.assertEqual(list(k(0, '1 2 3h')), [1, 2, 3])

    def test_int(self):
        x = [1, 2, 3, None]
        self.assertEqual(list(I(x)), x)

    def test_long(self):
        self.assertEqual(list(k(0, '1 2 3j')), [1, 2, 3])

    def test_date(self):
        self.assertEqual(list(k(0, '"d"$0 1 2')),
                         [date(2000, 1, d) for d in [1, 2, 3]])
        self.assertEqual(list(k(0, '-0W 0N 0Wd')),
                         [date(1, 1, 1), None, date(9999, 12, 31), ])

    def test_month(self):
        self.assertEqual(list(k(0, '"m"$0 1 2')),
                         [date(2000, m, 1) for m in [1, 2, 3]])
        self.assertEqual(list(k(0, '-0W 0N 0Wm')),
                         [date(1, 1, 1), None, date(9999, 12, 1), ])

    def test_datetime(self):
        self.assertEqual(list(k(0, '"z"$0.5 1.5 2.5')),
                         [datetime(2000, 1, d, 12) for d in [1, 2, 3]])
        self.assertEqual(list(k(0, '-0w 0n 0wz')), [datetime(1, 1, 1), None,
                                                    datetime(9999, 12, 31, 23,
                                                             59, 59,
                                                             999999), ])

    def test_time(self):
        self.assertEqual(list(k(0, '"t"$0 1 2 0N')),
                         [time(0, 0, 0, m * 1000) for m in range(3)] + [None])

    def test_minute(self):
        self.assertEqual(list(k(0, '"u"$0 1 2 0N')),
                         [time(0, m, 0) for m in range(3)] + [None])

    def test_second(self):
        self.assertEqual(list(k(0, '"v"$0 1 2')),
                         [time(0, 0, s) for s in range(3)])

    def test_symbol(self):
        x = ['a', 'b', 'c']
        self.assertEqual(list(S(x)), x)

    def test_float(self):
        x = [1.2, 2.3, 3.4]
        self.assertEqual(list(F(x)), x)

    def test_generic(self):
        x = kf(1.2), ki(2), ks('xyz')
        self.assertEqual([kstr(i) for i in K(x)],
                         [kstr(i) for i in x])

    def test_dict(self):
        d = k(0, '`a`b`c!1 2 3')
        self.assertEqual(list(d), list('abc'))

    def test_table(self):
        t = k(0, '([]a:`X`Y`Z;b:"xyz";c:1 2 3)')
        d = ['`a`b`c!(`X;"x";1)',
             '`a`b`c!(`Y;"y";2)',
             '`a`b`c!(`Z;"z";3)', ]
        for x, y in zip(t, d):
            self.assert_k_is(x, y)

    if Q_VERSION >= 3:
        def test_guid(self):
            s = '0x16151413121110090807060504030201'
            x = int(s, 16)
            self.assert_k_is(kguid(x), '0x0 sv ' + s)
            self.assertEqual([x], list(k(0, 'enlist 0x0 sv ' + s)))


class CallsTestCase(K_TestCase):
    def test_dot(self):
        x = k(0, '+')._dot(I([1, 2]))
        y = '3i'
        self.assert_k_is(x, y)
        self.assertRaises(_k.error, q('+')._dot, q('``'))

    def test_a0(self):
        x = k(0, '{1}')._a0()
        self.assert_k_is(x, '1')

    def test_a1(self):
        x = k(0, 'neg')._a1(ki(1))
        y = '-1i'
        self.assert_k_is(x, y)

    def test_call(self):
        f = q('{5}')
        x = ki(42)
        self.assert_k_is(f(), '5')
        self.assert_k_is(f(x), '5')
        self.assertRaises(_k.error, f, x, x)


class StrTestCase(K_TestCase):
    def test_pass(self):
        x = 'abc'
        self.assertEqual(str(ks(x)), x)
        self.assertEqual(str(kp(x)), x)
        x = b'a'
        self.assertEqual(str(kc(x)), x.decode())

    def test_misc(self):
        self.assertEqual(str(kb(1)), '1b')
        self.assertEqual(str(kh(1)), '1h')
        if Q_VERSION >= 3:
            self.assertEqual(str(ki(1)), '1i')
            self.assertEqual(str(kj(1)), '1')
        else:
            self.assertEqual(str(ki(1)), '1')
            self.assertEqual(str(kj(1)), '1j')
        self.assertEqual(str(kf(1)), '1f')


class ReprTestCase(unittest.TestCase):
    def test_pass(self):
        x = 'abc'
        self.assertEqual(repr(ks(x)), "k('`abc')")
        x = b'abc'
        self.assertEqual(repr(kp(x)), "k('\"abc\"')")
        x = b'a'
        self.assertEqual(repr(kc(x)), "k('\"a\"')")

    def test_misc(self):
        self.assertEqual(repr(kb(1)), "k('1b')")
        self.assertEqual(repr(kh(1)), "k('1h')")
        if Q_VERSION >= 3:
            self.assertEqual(repr(ki(1)), "k('1i')")
            self.assertEqual(repr(kj(1)), "k('1')")
        else:
            self.assertEqual(repr(ki(1)), "k('1')")
            self.assertEqual(repr(kj(1)), "k('1j')")
        self.assertEqual(repr(kf(1)), "k('1f')")


class JoinTestCase(K_TestCase):
    def test_any(self):
        x = ktn(0, 0)
        x._ja(ktj(101, 0))
        x._ja(ktj(101, 0))
        self.assert_k_is(x, '(::; ::)')

    def test_bool(self):
        x = q('10b')
        x._ja(True)
        y = '101b'
        self.assert_k_is(x, y)

    def test_byte(self):
        x = q('0x0102')
        x._ja(3)
        y = '0x010203'
        self.assert_k_is(x, y)

    def test_short(self):
        x = q('1 2h')
        x._ja(3)
        x._ja(-1)
        y = '1 2 3 -1h'
        self.assert_k_is(x, y)

    def test_int(self):
        x = q('1 2')
        x._ja(3)
        x._ja(-1)
        y = '1 2 3 -1'
        self.assert_k_is(x, y)

    def test_long(self):
        x = q('1 2j')
        x._ja(3)
        x._ja(-1)
        y = '1 2 3 -1j'
        self.assert_k_is(x, y)

    def test_real(self):
        x = q('1 2e')
        x._ja(3)
        x._ja(-1)
        y = '1 2 3 -1e'
        self.assert_k_is(x, y)

    def test_float(self):
        x = q('1 2f')
        x._ja(3)
        x._ja(-1)
        y = '1 2 3 -1f'
        self.assert_k_is(x, y)

    def test_str(self):
        x = q('"ab"')
        x._ja(b'c')
        y = '"abc"'
        self.assert_k_is(x, y)

    def test_sym(self):
        x = q('`a`b')
        x._ja(b'c')
        y = '`a`b`c'
        self.assert_k_is(x, y)


@pytest.mark.parametrize('f', [
    B, I, J, F, S, K,
])
def test_ja_errors(f):
    x = f([])
    with pytest.raises(TypeError):
        x._ja({})


@pytest.mark.parametrize('x', [
    "0x0102", "0 0h", "0 0e", '""',
])
def test_ja_errors_2(x):
    x = q(x)
    with pytest.raises(TypeError):
        x._ja({})


def test_ja_errors_3():
    x = kp("")
    with pytest.raises(TypeError):
        x._ja(b'aa')
    x = q('0x0102')
    with pytest.raises(OverflowError):
        x._ja(258)
    with pytest.raises(OverflowError):
        x._ja(-1)
    x = q('::')
    with pytest.raises(NotImplementedError):
        x._ja(0)


@pytest.mark.parametrize('f', [
    B, I, J, F, S, K,
])
def test_ja_overflow_errors(f):
    x = f([])
    with pytest.raises(TypeError):
        x._ja([])


@pytest.mark.parametrize('t', [
    'h', 'i', 'j', 'e', 'f'
])
def test_ja_none(t):
    x = q('"%s"$()' % t)
    x._ja(None)
    assert eq(x, q('enlist 0N' + t))


@pytest.mark.parametrize('t, x, y', [
    ('h', 2 ** 16, '0Wh'),
    ('i', 2 ** 32, '0Wi'),
    ('j', 2 ** 64, '0Wj'),
    ('h', - 2 ** 16, '-0Wh'),
    ('i', - 2 ** 32, '-0Wi'),
    ('j', - 2 ** 64, '-0Wj'),
    ('e', 1e40, '0We'),
    ('e', -1e40, '-0We'),
])
def test_ja_inf(t, x, y):
    a = q('"%s"$()' % t)
    a._ja(x)
    assert eq(a, q('enlist ' + y))


class ErrorTestCase(unittest.TestCase):
    def test_simple(self):
        self.assertRaises(_k.error, q, "1+`")
        self.assertRaises(_k.error, q("1+"), ks(''))
        self.assertRaises(_k.error, q("+"), ki(1), ks(''))

    def test_nested(self):
        self.assertRaises(_k.error, q, "{{{'`xyz}0}0}", ki(0))


class ArrayStructTestCase(unittest.TestCase):
    def test_type(self):
        for x in [q('1 2 3'), q('1')]:
            s = x.__array_struct__
            if _PY3K:
                self.assertEqual(type(s).__name__, 'PyCapsule')
            else:
                self.assertEqual(type(s).__name__, 'PyCObject')

    def test_error(self):
        x = q('()!()')
        self.assertRaises(AttributeError, lambda: x.__array_struct__)


class SerializationTestCase(K_TestCase):
    def test_b9(self):
        self.assert_k_is(b9(0, kb(1)), '0x010000000a000000ff01')
        self.assert_k_is(b9(0, kc(b'a')), '0x010000000a000000f661')
        self.assert_k_is(b9(0, kg(1)), '0x010000000a000000fc01')
        self.assert_k_is(b9(0, ki(1)), '0x010000000d000000fa01000000')
        self.assert_k_is(b9(0, kj(1)), '0x0100000011000000f90100000000000000')
        self.assert_k_is(b9(0, kp(b'')), '0x010000000e0000000a0000000000')

    def test_d9(self):
        self.assert_k_is(d9(q('0x010000000a000000ff00')), "0b")


try:
    from numpy import array
except ImportError:
    pass
else:
    class NumPyTestCase(unittest.TestCase):
        def test_scalar(self):
            x = ki(0)
            self.assertEqual(array(x).shape, ())


def eq(a, b):
    return q('~', a, b).inspect(b'g')


def test_B():
    x = B([])
    assert eq(x, q('0#0b'))

    x = B([True])
    assert eq(x, q('1#1b'))

    x = B([True, False])
    assert eq(x, q('10b'))

    with pytest.raises(TypeError):
        B(True, '')


@pytest.mark.parametrize("k,x", [
    ('1 2', [1, 2]),
    ('`a`b', ['a', 'b']),
    ('1 2e', [1., 2.]),
])
def test_as_sequence(k, x):
    a = q(k)
    b = list(a)
    assert len(a) == len(b)
    assert a[0] == b[0]
    assert b == x


@pytest.mark.parametrize("k,n", [
    ('1 2', 2),
    ('`a`b', 2),
    ('`a`b!1 2', 2),
    ('([]a:1 2;b:0;c:0)', 2),
    ('([]a:1 2)', 2),
    ('([a:1 2]b:10 20)', 2),
    ('0#([a:1 2]b:10 20)', 0),
    ('0', 1),
    ('{}', 1),
])
def test_length(k, n):
    assert len(q(k)) == n


def test_k_methods():
    assert _k.ymd(2000, 1, 1) == 0
    assert _k.ymd(1970, 1, 1) == -10957
    assert _k.dj(0) == 20000101
    assert _k.dj(-10957) == 19700101


def test_jv():
    x = q('1 2')
    y = q('3 4')
    x._jv(y)
    assert eq(x, q('1 2 3 4'))
    assert y.inspect(b'r') == 0
    with pytest.raises(TypeError):
        x._jv(1)


@pytest.mark.parametrize("k,py", [
    ('-0Wd', date(1, 1, 1)),
    ('0Nd', None),
    ('0Wd', date(9999, 12, 31)),
    ('2014.06.06', date(2014, 6, 6)),
    ('1828.09.09', date(1828, 9, 9)),
])
def test_date(k, py):
    x = q('enlist ' + k)
    assert x[0] == py


def test_errors_type():
    with pytest.raises(_k.error) as e:
        q('`+1')
    assert e.value.args[0] == 'type'

    with pytest.raises(TypeError):
        ktd(None)

    with pytest.raises(TypeError):
        ka(None, None)

    with pytest.raises(TypeError):
        dot(None, None)

    with pytest.raises(TypeError):
        a1(None, None)

    with pytest.raises(TypeError):
        ktn(None, None)

    # with pytest.raises(TypeError):
    #     q('{}')(None)

    with pytest.raises(TypeError):
        k(None)


@pytest.mark.parametrize('f', [
    G, H, I, J, E, F, S, K, P, M, D, N, U, V, T,
])
def test_sequence_conversion_errors(f):
    with pytest.raises(TypeError):
        f(None)
    with pytest.raises(TypeError):
        f([()])


@pytest.mark.parametrize('f', [
    ka, kg, kh, ki, kj, ke, kf, kc, ks, km, kd, kz, ku, kv, kt, kp,
    kzz, knz, kpz, b9, d9,
])
def test_scalar_conversion_errors(f):
    with pytest.raises(TypeError):
        f([])


@pytest.mark.parametrize('f,x', [
    (G, 1), (I, 1), (J, 1), (F, 1.0), (S, ''),
])
def test_sequence_mixed_type_errors(f, x):
    with pytest.raises(TypeError):
        f([x, ()])


def test_errors_ktd():
    with pytest.raises(TypeError):
        ktd(None)
    x = ki(0)
    with pytest.raises(_k.error):
        ktd(x)
    # Check that ktd does not leak refs of failure.
    assert x._r == 0


def test_errors_ka():
    with pytest.raises(TypeError):
        ka(None, None)


def test_errors_dot():
    with pytest.raises(TypeError):
        dot(None, None)
    f = ktj(101, 0)  # ::
    with pytest.raises(TypeError):
        dot(f, None)


def test_errors_a1():
    with pytest.raises(TypeError):
        a1(None, None)


def test_errors_ktn():
    with pytest.raises(TypeError):
        ktn(None, None)


# def test_errors_q_arg0():
#     with pytest.raises(TypeError):
#         q('{}')(None)


def test_errors_k():
    with pytest.raises(TypeError):
        k(None)
    with pytest.raises(TypeError):
        k(0, None)
    with pytest.raises(TypeError):
        k(0, ':', None)


def test_errors_K():
    with pytest.raises(TypeError):
        _k.K()


def test_errors_ymd():
    with pytest.raises(TypeError):
        _k.ymd()


def test_errors_dj():
    with pytest.raises(TypeError):
        _k.dj()


def test_errors_I():
    assert eq(I([2 ** 32]), q('1#0Wi'))
    assert eq(I([2 ** 64]), q('1#0Wi'))


def test_errors_J():
    assert eq(J([2 ** 64]), q('1#0Wj'))
    assert eq(J([2 ** 64 + 1]), q('1#0Wj'))


@pytest.mark.parametrize('f', [
    ka, kg, kh, ki, kj, ke, kf, kc, ks, km, kd, kz, ku, kv, kt, kp,
    kzz, knz, kpz, b9, d9, xD, xT, kzz,
] + ([kguid] if Q_VERSION >= 3 else []))
def test_misc_conversion_errors(f):
    with pytest.raises(TypeError):
        f([])


def test_reference_leaks():
    x = q('1 2 3')
    n = q('\\w')[0]
    s = str(x)
    assert n == q('\\w')[0]
    r = repr(x)
    assert n == q('\\w')[0]
    test_errors_I()
    assert n == q('\\w')[0]
    test_errors_J()
    assert n == q('\\w')[0]


def test_ktd_reference_leak():
    x = q('([a:1 2]b:3 4)')
    ktd(x)
    assert x.inspect(b'r') == 0


def test_xD_reference_leak():
    x = q('1 2 3')
    y = q('5 6 7')
    xD(x, y)
    assert x.inspect(b'r') == 0
    assert y.inspect(b'r') == 0


def test_func():
    def f(x, y):
        return y

    kf = func(f)
    x = q('1 2j')

    assert eq(kf(x), kj(2))


@pytest.mark.parametrize('x,r', [
    (1, '1'),
    (3.14, '3.14'),
    (None, '::'),
    (True, '1b'),
    (False, '0b')
])
def test_func_values(x, r):
    def f():
        return x

    kf = func(f)
    assert eq(kf(ktn(0, 0)), q(r))


@pytest.mark.parametrize('x,e', [
    ([], 'py2k'),
    (2 ** 65, 'OverflowError'),
])
def test_func_errors(x, e):
    def f():
        return x

    kf = func(f)
    with pytest.raises(_k.error) as info:
        kf(ktn(0, 0))
        assert info.value.args[0] == e


def test_ktn():
    x = ktn(_k.KF, 3)
    assert x.inspect(b't') == _k.KF
    assert x.inspect(b'n') == 3


@pytest.mark.parametrize("k,py", [
    ('::', None),
    ('1b', True),
    ('0b', False),
    ('42', 42),
    ('3.14', 3.14),
    ('2001.01.01', date(2001, 1, 1)),
    ('2001.01.01D0', datetime(2001, 1, 1)),
    ('0D01:02:03.456789',
     timedelta(hours=1, minutes=2, seconds=3, microseconds=456789)),
    ('`xyz', 'xyz'),
])
def test_constructor(k, py):
    x = _k.K(py)
    assert eq(q(k), x)


@pytest.mark.skipif("sys.version_info >= (3,0)")
def test_constructor_long():
    x = eval('42L')
    assert eq(_k.K(x), q('42'))


@pytest.mark.parametrize('x', list('zmtuvd'))  # nyi - np
def test_temporal_none(x):
    assert q('enlist 0N' + x)[0] is None


def test_inspect():
    x = q("`s#1 2")
    assert x.inspect(b'u') == 1


def test_enlist():
    for n in range(1, 8):
        args = [kc(b'x')] + [ki(0)] * n
        x = q('enlist', *args)
        y = knk(len(args), *args)
        # TODO: figure out why simple == comparison fails
        assert str(x) == str(y)
        errargs = [None] + args[1:]
        with pytest.raises(TypeError):
            q('enlist', *errargs)
        with pytest.raises(TypeError):
            knk(len(errargs), *errargs)


@pytest.mark.parametrize('x, y', [
    ('"abc"', '"xxx"'),
    ('1 2 3h', '0 0 0h'),
    ('1 2 3i', '0 0 0i'),
    ('1 2 3', '0 0 0'),
])
def test_buffer_protocol(tmpdir, x, y):
    x = q(x)
    y = q(y)
    p = tmpdir.join('x')
    with p.open('wb') as f:
        f.write(x)
    with p.open() as f:
        if _PY3K:
            f = f.buffer
        f.readinto(y)

    assert eq(x, y)


def test_charbuffer(tmpdir):
    x = q('"abc"')
    p = tmpdir.join('x')
    with p.open('w') as f:
        if _PY3K:
            f = f.buffer
        f.write(x)

    assert p.read() == 'abc'


def test_charbuffer_error(tmpdir):
    x = q('1 2 3j')
    p = tmpdir.join('x')
    with p.open('w') as f:
        if _PY3K:
            f = f.buffer
            f.write(x)
            f.close()
            assert struct.unpack("qqq", p.read_binary()) == (1, 2, 3)
        else:
            with pytest.raises(TypeError):
                f.write(x)


def test_iter():
    x = q('0')
    with pytest.raises(TypeError):
        iter(x)

    x = q('1 2')
    i = iter(x)
    assert list(i) == [1, 2]


@pytest.mark.parametrize('Y, M, D, h, m, s, u', [
    (1970, 1, 1, 0, 0, 0, 123000),
    (2015, 1, 13, 20, 27, 25, 926000),
])
def test_datetime_conversion(Y, M, D, h, m, s, u):
    d = datetime(Y, M, D, h, m, s, u)
    x = q('enlist', kzz(d))
    assert x[0] == d


def test_errmsg():  # Issue #737
    with pytest.raises(ValueError) as e:
        F([1., ''])
    assert 'float' in e.value.args[0]


@pytest.mark.parametrize('x,y', [
    ('0b', False),
    ('1b', True),
    ('"x"', 'x'),
    ('0x00', 0),
    ('0h', 0),
    ('0i', 0),
    ('0', 0),
    ('0e', 0.0),
    ('0f', 0.0),
    ('0Wj-1', 9223372036854775806),
])
def test_pys(x, y):
    k = q(x)
    s = k._pys()
    assert s == y
    assert type(s) is type(y)


def test_pys_vector():
    x = q('1 2')
    with pytest.raises(TypeError):
        x._pys()


def test_pys_nyi():
    # Timestamp conversion is not implemented to
    # avoid the loss of precision.
    x = ka(-_k.KP, 0)
    with pytest.raises(NotImplementedError):
        x._pys()


def test_getitem_neg():
    x = I([0, 1, 2, 3])
    assert x[-2] == 2


@pytest.mark.parametrize('i', [-5, 4, 10])
def test_getitem_out_of_range(i):
    x = I([0, 1, 2, 3])
    with pytest.raises(IndexError):
        x[i]


def test_getitem_nyi():
    x = q('0N 0Np')
    with pytest.raises(NotImplementedError):
        x[0]


@pytest.mark.parametrize('x', "hijmdvut")
def test_getitem_none(x):
    x = q('enlist 0N' + x)
    assert x[0] is None


# XXX: This is commented out because it induces failures
# in unrelated tests.  TODO: investigate whether additional
# cleanup is needed upon a connection error.
# def test_connection_error():
#    with pytest.raises(OSError):
#        k(-1, "")

def test_issue_796():
    x = J([1])
    assert type(x[0]) is int


@pytest.mark.skipif('_PY3K')
def test_issue_796_long():
    maxint = sys.maxint
    if maxint == 2 ** 31 - 1:
        x = J([maxint, -maxint - 1, maxint + 1, -maxint - 2])
        assert [type(i) for i in x] == [int, int, long, long]
    else:
        x = J([maxint - 1, -maxint + 1])
        assert [type(i) for i in x] == [int, int]


def test_okx():
    x = kb(0)
    b = b9(0, x)
    assert _k.okx(b)


@pytest.mark.parametrize('b', [None, kb(0)])
def test_okx_errors(b):
    with pytest.raises(TypeError):
        _k.okx(b)


@pytest.mark.parametrize('attr,value', [
    ('r', 0),
    ('t', 100),
])
def test_int_attr(attr, value):
    x = q('{}')
    a = getattr(x, '_' + attr)
    assert isinstance(a, int)
    assert a == value


# Atoms and vectors
# q) .Q.t
# " bg xhijefcspmdznuvts"
# num ch name       atom    vector
# --------------------------------
# 1   b  boolean    kb       B
# 2   g  guid       kguid    -
# 4   x  byte       kg       -
# 5   h  short      kh       -
# 6   i  int        ki       I
# 7   j  long       kj       J
# 8   e  real       ke       -
# 9   f  float      kf       F
# 10  c  char       kc       kp
# 11  s  symbol     ks       S
# 12  p  timestamp  knz      -
# 13  m  month      km       -
# 14  d  date       kd       D
# 15  z  datetime   kz       -
# 16  n  timespan   kpz      -
# 17  u  minute     ku       -
# 18  v  second     kv       -
# 19  t  time       kt       -
# 11  s  symbol     ks       -

def test_byte_list():
    x = G([])
    assert x._t == _k.KG
    x = G([None, 1])
    assert eq(x, q('0x0001'))


@pytest.mark.parametrize('ch, num', [
    ("H", 5),
    ("I", 6),
    ("J", 7),
])
def test_integer_list(ch, num):
    vec = getattr(_k.K, '_%s' % ch)
    # Empty list - check the type
    x = vec([])
    assert x._t == num
    # List with a missing value
    x = vec([None, 0])
    assert eq(x, q('0N 0' + ch.lower()))


@pytest.mark.parametrize('ch, num', [
    ("E", 8),
    ("F", 9),
])
def test_float_list(ch, num):
    vec = getattr(_k.K, '_%s' % ch)
    # Empty list - check the type
    x = vec([])
    assert x.inspect(b't') == num
    # List with special values
    inf = float('inf')
    nan = float('nan')
    x = vec([0.0, -0.0, inf, -inf, nan])
    assert x[0] == 0.0
    assert math.copysign(1.0, x[1]) < 0
    assert math.isinf(x[2]) and x[2] > 0
    assert math.isinf(x[3]) and x[3] < 0
    assert math.isnan(x[4])


def test_timestamp_list():
    x = P([])
    assert x._t == _k.KP
    d = date(2000, 1, 1)
    x = P([None, d])
    assert eq(x, q('0N 2000.01.01D0'))
    dt = datetime(2000, 1, 1, 12, 34, 56, 789)
    x = P([None, dt])
    assert eq(x, q('0N 2000.01.01D12:34:56.000789'))


def test_month_list():
    x = M([])
    assert x._t == _k.KM
    d = date(2000, 1, 1)
    x = M([None, d])
    assert eq(x, q('0N 2000.01m'))
    dt = datetime(2000, 1, 1, 12, 34, 56, 789)
    x = M([None, dt])
    assert eq(x, q('0N 2000.01m'))


def test_timespan_list():
    x = N([])
    assert x._t == _k.KN
    d = timedelta(1, 2, 3)
    x = N([None, d, timedelta.max, timedelta.min])
    assert eq(x, q('0N 1D00:00:02.000003 0W -0Wn'))


def test_minute_list():
    x = U([])
    assert x._t == _k.KU
    t = time(1, 2, 3, 4000)
    x = U([None, t])
    assert eq(x, q('0N 01:02'))


def test_second_list():
    x = V([])
    assert x._t == _k.KV
    t = time(1, 2, 3, 4000)
    x = V([None, t])
    assert eq(x, q('0N 01:02:03'))


def test_time_list():
    x = T([])
    assert x._t == _k.KT
    t = time(1, 2, 3, 4000)
    x = T([None, t])
    assert eq(x, q('0N 01:02:03.004'))


@pytest.mark.skipif("Q_VERSION < 3")
def test_guid_list():
    x = UU([])
    assert x._t == _k.UU
    x = UU([None, 0x01020304050607080910111213141516])
    assert str(x) == ('00000000-0000-0000-0000-000000000000 '
                      '01020304-0506-0708-0910-111213141516')


@pytest.mark.parametrize("t", "bxhijefcnuvt")
def test_false_scalar(t):
    x = q('"%c"$0' % t)
    assert not x


@pytest.mark.parametrize("t", "bxhijefcpmdznuvt")
def test_true_scalar(t):
    x = q('"%c"$1' % t)
    assert x


@pytest.mark.parametrize("x,r", [
    ('`', False),
    ('`a', True),
    ('::', False),
    ('{}', True),
    ('+', True),
    ('neg', True),
    ('raze', True),
    ('()!()', False),
    ('1 2!3 4', True),
    ('flip`a`b!()', False),
    ('([]a:1 2)', True),
    ('([k:1 2]a:1 2)', True),
    ('1!flip`a`b!()', False),
])
def test_k_bool(x, r):
    x = q(x)
    assert r == bool(x)


def test_k_bool_of_enum():
    q("sym:`a`b`c")
    x = q("`sym$`a")
    assert x
    x = q("`sym$`")
    assert not x


def test_enum_getitem():
    q("sym:`a`b`c")
    x = q('`sym$`a`')
    assert x[0] == 'a'
    assert x[1] == ''
    with pytest.raises(IndexError):
        x[2]
    # Test short enum domain vector
    q("sym:1#sym")
    assert x[1] == ''


@pytest.mark.skipif("Q_VERSION < 3")
def test_setm():
    assert _k.setm(1) == 0
    assert _k.setm(0) == 1
    assert _k.setm(0) == 0
    with pytest.raises(TypeError):
        _k.setm('')


@pytest.mark.skipif("Q_VERSION < 3")
def test_m9():
    # For now just check that m9() exists.
    # Note that m9() can only be called from
    # a separate thread before its termination.
    assert _k.m9


@pytest.mark.parametrize("x, n", [
    ("()", 0),
    ("1 2", 2),
    ("()!()", 2),
])
def test_attr_n(x, n):
    x = q(x)
    assert n == x._n


@pytest.mark.skipif("Q_VERSION >= 3.6")
def test_date_range():
    min_date = date(1709, 1, 1)
    max_date = date(2290, 12, 31)
    one_day = timedelta(1)
    result = q('-0W 1709.01.01 1709.01.02 2290.12.30 2290.12.31 0Wd')
    assert eq(D([min_date - one_day, min_date, min_date + one_day,
                 max_date - one_day, max_date, max_date + one_day]),
              result)
    # Demonstrate that q bails on the out of range dates
    with pytest.raises(_k.error):
        d1 = min_date - one_day
        # We cannot use strftime here because in Python 2.7 it only
        # works for dates after 1900.
        q("%d.%d.%d" % (d1.year, d1.month, d1.day))
    with pytest.raises(_k.error):
        d2 = max_date + one_day
        q(d2.strftime('%Y.%m.%d'))

    # Test scalar conversions
    assert eq(_k.K(d1), q('-0Wd'))
    assert eq(_k.K(d2), q('0Wd'))
    assert eq(_k.K(min_date), q('1709.01.01'))
    assert eq(_k.K(max_date), q('2290.12.31'))


def test_issue_870():
    datetimes = [datetime(2100, 1, 1), datetime(2200, 1, 1)]
    result = q('2100.01.01D00:00:00.000000000 2200.01.01D00:00:00.000000000')
    assert eq(P(datetimes), result)


@pytest.mark.skipif("Q_VERSION >= 3.6")
def test_month_range():
    dates = map(date,
                [1708, 1709, 2290, 2291],
                [12, 1, 12, 1], [1] * 4)
    result = q('-0W 1709.01 2290.12 0Wm')
    assert eq(M(dates), result)


def test_timestamp_range():
    min_datetime = datetime(1709, 1, 1)
    over_datetime = datetime(2291, 1, 1)
    eps = datetime.resolution
    datetimes = [min_datetime - eps, min_datetime,
                 over_datetime - eps, over_datetime]
    result = q(
        '-0W 1709.01.01D00:00:00.000000000 2290.12.31D23:59:59.999999000 0Wp')
    assert eq(P(datetimes), result)


@pytest.mark.skipif("Q_VERSION < 3.6")
def test_date_and_month_extremes():
    dates = [date.min, date.max]
    assert eq(M(dates), q('0001.01 9999.12m'))
    assert eq(D(dates), q('0001.01.01 9999.12.31'))


@pytest.mark.parametrize("t,f", [
    ("boolean", "?"),
    ("byte", "B"),
    ("short", "h"),
    ("int", K_INT_CODE),
    ("long", K_LONG_CODE),
    ("real", "f"),
    ("float", "d"),
    ("char", "s"),
    ("symbol", "P"),
    ("timestamp", K_LONG_CODE),
    ("timespan", K_LONG_CODE),
    ("date", K_INT_CODE),
    ("minute", K_INT_CODE),
    ("second", K_INT_CODE),
    ("datetime", "d"),
])
def test_data_attribute(t, f):
    x = q('$', ks(t), ktn(0, 0))
    assert x.data.format == f


def test_data_attribute_error():
    x = q('::')
    with pytest.raises(AttributeError):
        x.data


def test_n_attribute_error():
    x = ki(0)
    with pytest.raises(AttributeError):
        x._n


@pytest.mark.parametrize('x,m,r', [
    ('1j', 't', -7),
    # pytest.mark.skipif("Q_VERSION < 3", ('1', 'a', 3)),
    # -- unpredictable values
    pytest.param('1', 'm', 0, marks=pytest.mark.skipif("Q_VERSION < 3")),
    ('"x"', 'c', b'x'),
    ('enlist 0x42', 'G', 0x42),
    ('enlist 42h', 'H', 42),
    ('enlist 42i', 'I', 42),
    ('enlist 42j', 'J', 42),
    ('enlist 42e', 'E', 42.),
    ('enlist 42f', 'F', 42.),
])
def test_inspect(x, m, r):
    x = q(x)
    if m.isupper():
        assert x.inspect(m, 0) == r
    else:
        assert x.inspect(m) == r


def test_inspect_error():
    x = ki(0)
    with pytest.raises(TypeError):
        x.inspect(42)


def test_a1_error():
    x = ki(0)
    with pytest.raises(TypeError):
        x._a1(None)


def test_str_error():
    x = q('`sym?`x')
    value = q('value')
    q(".q.value:{'`error}")
    try:
        with pytest.raises(_k.error):
            str(x)
    finally:
        q('set', ks('.q.value'), value)


def test_call_kw_error():
    f = q('::')
    with pytest.raises(TypeError):
        f(a=None)
    f = q('{[a;b]a-b}')
    with pytest.raises(TypeError):
        f(1, a=2)
    with pytest.raises(TypeError):
        f(1, 2, z=3)


def test_data_attr_errors():
    # Empty list
    x = ktn(0, 0)
    with pytest.raises(AttributeError):
        x.data
    # Scalar in generic list
    x = knk(2, ki(0), kj(0))
    with pytest.raises(AttributeError):
        x.data
    # Size varies
    x = knk(2, I([]), I([0]))
    with pytest.raises(AttributeError):
        x.data


def test_kb_error():
    class X:
        def __index__(self):
            raise ValueError

    x = X()
    with pytest.raises(ValueError):
        kb(x)
    with pytest.raises(ValueError):
        B([x])

    with pytest.raises(TypeError):
        kb(0.0)


@pytest.mark.parametrize("f, u, o, r", [
    ('kg', -1, 256, OverflowError),
    ('kh', -2 ** 15 - 1, 2 ** 15 + 1, '0Wh'),
    ('ki', -2 ** 31 - 1, 2 ** 31 + 1, '0Wi'),
    ('ke', -1e200, 1e200, '0We'),
    ('kf', -float('inf'), float('inf'), '0w'),
    ('kd', -2 ** 31 - 1, 2 ** 31 + 1, '0Wd'),
    pytest.param('kd', date.min, date.max, '0Wd',
                 marks=pytest.mark.skipif('Q_VERSION >= 3.6')),
    ('km', -2 ** 31 - 1, 2 ** 31 + 1, '0Wm'),
    ('ku', -2 ** 31 - 1, 2 ** 31 + 1, '0Wu'),
    ('kv', -2 ** 31 - 1, 2 ** 31 + 1, '0Wv'),
    ('kpz', -2 ** 63 - 1, 2 ** 63 + 1, '0Wp'),
    ('knz', -2 ** 63 - 1, 2 ** 63 + 1, '0Wn'),
])
def test_scalar_overflow(f, u, o, r):
    f = getattr(_k.K, '_' + f)
    if isinstance(r, type):
        for v in [u, o]:
            with pytest.raises(r):
                f(v)
    else:
        s = q('-' + r)
        r = q(r)
        assert eq(f(u), s)
        assert eq(f(o), r)


def test_repr():
    assert repr(ktj(101, 1)) == "k('+:')"


def test_kj_error():
    class X:
        def __index__(self):
            return ''

    x = X()
    with pytest.raises(TypeError):
        kj(x)


def test_kc():
    assert eq(kc('a'), q('"a"'))
    assert eq(kc(b'a'), q('"a"'))
    with pytest.raises(TypeError):
        kc({})
    with pytest.raises(TypeError):
        kc('ab')
    with pytest.raises(TypeError):
        kc(b'ab')


@pytest.mark.skipif("Q_VERSION < 3.5")
def test_trp():
    f = q('{x+y}')
    args = q('(1;`)')
    with pytest.raises(_k.error) as info:
        f._trp(args)
    a = info.value.args
    assert a[0] == 'type'
    bt = q('sublist', q('0 1'), a[1])
    sbt = q('.Q.sbt', bt)
    assert str(sbt) == (
        '  [3]  {x+y}\n'
        '         ^\n').replace('\n', os.linesep)


def test_callargs():
    f = q('{x+y}')
    assert f._callargs(1, y=2) == (1, 2)


@pytest.mark.parametrize('x', ['0Wt', '-0Wt'])
def test_inf_time_pys(x):
    x = q(x)
    with pytest.raises(OverflowError):
        x._pys()


@pytest.mark.skipif("Q_VERSION < 3.5")
def test_knt():
    t = q('([]a:1 2;b:3 4)')
    kt = q('([a:1 2];b:3 4)')
    assert eq(knt(1, t), kt)
    with pytest.raises(ValueError):
        knt(0, t)
    with pytest.raises(_k.error):
        knt(2, t)


def test_none_in_float_list():
    a = [1., None]
    b = F(a)
    assert eq(b, q('1 0n'))
    b = E(a)
    assert eq(b, q('1 0ne'))



"""REPL based on prompt-toolkit"""
import sys
import os

try:
    import ptpython.entry_points.run_ptpython as ptp
except ImportError:
    if hasattr(sys, '_called_from_test'):
        raise
    print('Cannot import ptpython. Try',
          '    pip install ptpython', sep='\n')
    raise SystemExit(1)

from pyq import q, kerr


def console_size(fd=1):
    """Return console size as a (LINES, COLUMNS) tuple"""
    try:
        import fcntl
        import termios
        import struct
    except ImportError:
        size = os.getenv('LINES', 25), os.getenv('COLUMNS', 80)
    else:
        size = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ,
                                               b'1234'))
    return size


def run(q_prompt=False):
    """Run a prompt-toolkit based REPL"""
    lines, columns = console_size()
    q(r'\c %d %d' % (lines, columns))
    if len(sys.argv) > 1:
        try:
            q(r'\l %s' % sys.argv[1])
        except kerr as e:
            print(e)
            raise SystemExit(1)
        else:
            del sys.argv[1]
    if q_prompt:
        q()
    ptp.run()



from __future__ import absolute_import
import sys
import pytest
import pkg_resources


def test_version():
    from pyq import __version__ as v
    pv = pkg_resources.parse_version(v)
    assert pv
    assert v != 'unknown'


@pytest.fixture
def no_version_pyq(monkeypatch):
    monkeypatch.setitem(sys.modules, 'pyq.version', None)
    monkeypatch.delitem(sys.modules, 'pyq', raising=False)
    monkeypatch.delitem(sys.modules, 'pyq._k', raising=False)
    import pyq
    return pyq


def test_no_version(no_version_pyq):
    assert no_version_pyq.__version__ == 'unknown'



from __future__ import absolute_import
from __future__ import unicode_literals
from ..magic import q as qmag, logical_lines


def test_q_magic_cell():
    r = qmag('', """\
a:1 2 3
b:3 2 1
a + b
""")
    assert r == [4] * 3


def test_q_magic_line():
    r = qmag('a:1 2 3;b:3 2 1;a + b')
    assert r == [4, 4, 4]


def test_logical_lines():
    script = """\
f:{a:x + y
   neg a}
f[5;6]
"""
    ll = logical_lines(script)
    assert next(ll) == "f:{a:x + y\n   neg a}"
    assert next(ll) == "f[5;6]"


def test_q_magic_cell_2():
    script = """\
f:{x +  / to be continued
 y}

f[5;6]
"""
    assert 11 == qmag('', script)


def test_q_magic_cell_3(tmpdir, monkeypatch):
    monkeypatch.chdir(tmpdir)
    script1 = tmpdir.join("script-1.q")
    script2 = tmpdir.join("script-2.q")
    script1.write("x:1")
    script2.write("x:2")

    x = qmag("-l script-1.q -l script-2.q", "x")
    assert x == 2


def test_q_magic_output(q):
    ns = {}
    qmag('-o x', "42", ns)
    assert ns['x'] == 42

    qmag('-o q.x', "42")
    assert q.x == 42


def test_q_magic_input():
    ns = dict(x=1, y=2)
    r = qmag('-i x,y', "x+y", ns)
    assert r == 3



from __future__ import absolute_import
import os

import pytest

from pyq import q


def test_inside_ci():
    # Test designed for Gitlab CI
    if 'GITLAB_CI' not in os.environ:
        pytest.skip("This test designed for Gitlab CI only.")

    qver, qmin = os.environ.get('KDB_VER', '0.0').split('.')
    assert q(".z.K") == int(qver) + 0.1 * int(qmin)



"""IPython magic"""
import io
import os
from tempfile import mkstemp
import pyq
from getopt import getopt
import sys

StringIO = io.StringIO  # LGTM.com does not like double import
STD_STREAM = [sys.stdin, sys.stdout, sys.stderr]

try:
    string_types = (str, unicode)
except NameError:
    string_types = (str,)

Q_NONE = pyq.q('::')


def logical_lines(lines):
    """Merge lines into chunks according to q rules"""
    if isinstance(lines, string_types):
        lines = StringIO(lines)
    buf = []
    for line in lines:
        if buf and not line.startswith(' '):
            chunk = ''.join(buf).strip()
            if chunk:
                yield chunk
            buf[:] = []

        buf.append(line)
    chunk = ''.join(buf).strip()
    if chunk:
        yield chunk


def _forward_outputs(outs):
    for fd in (1, 2):
        if fd in outs:
            os.lseek(fd, 0, os.SEEK_SET)
            with io.open(fd, closefd=False) as f:
                STD_STREAM[fd].writelines(f)
            os.ftruncate(fd, 0)


def q(line, cell=None, _ns=None):
    """Run q code.

    Options:
       -l (dir|script) - pre-load database or script
       -h host:port - execute on the given host
       -o var - send output to a variable named var.
       -i var1,..,varN - input variables
       -1/-2 - redirect stdout/stderr
    """
    if cell is None:
        return pyq.q(line)

    if _ns is None:
        _ns = vars(sys.modules['__main__'])
    input = output = None
    preload = []
    outs = {}
    try:
        h = pyq.q('0i')
        if line:
            for opt, value in getopt(line.split(), "h:l:o:i:12")[0]:
                if opt == '-l':
                    preload.append(value)
                elif opt == '-h':
                    h = pyq.K(str(':' + value))
                elif opt == '-o':
                    output = str(value)  # (see #673)
                elif opt == '-i':
                    input = str(value).split(',')
                elif opt in ('-1', '-2'):
                    outs[int(opt[1])] = None
        if outs:
            if int(h) != 0:
                raise ValueError("Cannot redirect remote std stream")
            for fd in outs:
                tmpfd, tmpfile = mkstemp()
                try:
                    pyq.q(r'\%d %s' % (fd, tmpfile))
                finally:
                    os.unlink(tmpfile)
                    os.close(tmpfd)
        r = None
        for script in preload:
            h(pyq.kp(r"\l " + script))
        if input is not None:
            for chunk in logical_lines(cell):
                func = "{[%s]%s}" % (';'.join(input), chunk)
                args = tuple(_ns[i] for i in input)
                if r != Q_NONE:
                    r.show()
                r = h((pyq.kp(func),) + args)
                if outs:
                    _forward_outputs(outs)
        else:
            for chunk in logical_lines(cell):
                if r != Q_NONE:
                    r.show()
                r = h(pyq.kp(chunk))
                if outs:
                    _forward_outputs(outs)
    except pyq.kerr as e:
        print("'%s" % e)
    else:
        if output is not None:
            if output.startswith('q.'):
                pyq.q('@[`.;;:;]', output[2:], r)
            else:
                _ns[output] = r
        else:
            if r != Q_NONE:
                return r


def _q_formatter(x, p, _):
    x_show = x.show(output=str)
    p.text(x_show.strip())


def load_ipython_extension(ipython):
    """Register %q and %%q magics and pretty display for K objects"""
    ipython.register_magic_function(q, 'line_cell')
    fmr = ipython.display_formatter.formatters['text/plain']
    fmr.for_type(pyq.K, _q_formatter)


def unload_ipython_extension(ipython):
    pass



from __future__ import absolute_import
from __future__ import unicode_literals
import subprocess
import os


def test_python_q_exitcode():
    with open(os.devnull) as rnull, open(os.devnull, 'w') as wnull:
        assert subprocess.call([os.getenv('QBIN'), 'python.q',
                                '-c@', 'raise Exception'],
                               stdin=rnull, stdout=wnull, stderr=wnull) == 1



from __future__ import absolute_import
import subprocess
import threading
from pyq import _k
import os
import sys


def test_m9_in_thread():
    def call_m9():
        _k.m9()

    thread = threading.Thread(target=call_m9)
    thread.start()
    thread.join()


if sys.version_info >= (3, ):
    MT_PEACH_CODE = "from threading import get_ident"
else:
    MT_PEACH_CODE = "from thread import get_ident"

MT_PEACH_CODE += """
q.f = get_ident
q)exit"i"$%d<>count distinct {f()} peach til 10
"""


def test_mt_peach(tmpdir):
    n = 4  # number of slave threads for peach
    qbin = os.getenv('QBIN')
    script = tmpdir.join('x.p')
    script.write(MT_PEACH_CODE % n)
    subprocess.check_call([qbin, script.strpath, '-s', str(n), '-q'])



from __future__ import absolute_import
import pytest
import os

try:
    import pathlib
except ImportError:
    try:
        import pathlib2 as pathlib
    except ImportError:
        pathlib = None
from pyq import K

pytestmark = pytest.mark.skipif(pathlib is None,
                                reason="pathlib is not available")


@pytest.mark.parametrize('x', [
    'a',
    os.path.join('a', 'b'),
    os.path.join('a', 'b', 'c'),
])
def test_convert_path(x):
    p = pathlib.Path(x)
    assert str(K(p)) == ':' + x.replace(os.sep, '/')



"""pytest fixtures and settings for pyq"""
import os

import pytest
import subprocess


def pytest_configure(config):
    """set pytest sentinel"""
    import sys
    sys._called_from_test = True


def pytest_unconfigure(config):
    """unset pytest sentinel"""
    import sys
    del sys._called_from_test


@pytest.fixture(autouse=True)
def add_array(doctest_namespace):
    """Allow using numpy array in doctests"""
    try:
        from numpy import array
    except ImportError:
        pass
    else:
        doctest_namespace['array'] = array


@pytest.fixture(autouse=True)
def add_dt(doctest_namespace):
    """Allow using datetime classes in doctests"""
    from datetime import datetime, date, time, timedelta
    doctest_namespace['datetime'] = datetime
    doctest_namespace['date'] = date
    doctest_namespace['time'] = time
    doctest_namespace['timedelta'] = timedelta


@pytest.fixture
def kdb_server(q, tmpdir):
    """Starts a kdb server and yields connection handle"""
    from pyq import kerr
    prog = os.getenv('QBIN')
    script = tmpdir.join('srv.q')
    script.write("""\
p:1024;while[0~@[system;"p ",string p;0];p+:1];-1 string p;
""")
    with open(os.devnull) as devnull:
        process = subprocess.Popen([prog, script.strpath],
                                   stdout=subprocess.PIPE,
                                   stdin=devnull)
        line = process.stdout.readline()
        port = int(line.strip())
        c = q.hopen(port)
        c(q('".z.pc:{exit 0}"'))  # stop server on port close
        try:
            yield c
        finally:
            try:
                c.hclose()
            except kerr:
                pass
            process.stdout.close()
            process.wait()



"""Prompt toolkit"""
import re

from prompt_toolkit.completion import Completion
from prompt_toolkit.contrib.completers import PathCompleter
from prompt_toolkit.contrib.completers.base import Completer
from prompt_toolkit.document import Document
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit import prompt
from prompt_toolkit.styles import style_from_pygments

from pygments.token import Token
from pygments.styles.monokai import MonokaiStyle as BasicStyle
from pygments.lexers import get_lexer_by_name

try:
    q_lexer = get_lexer_by_name('q')
except ValueError:
    lexer = None
else:
    from prompt_toolkit.layout.lexers import PygmentsLexer

    lexer = PygmentsLexer(type(q_lexer))
from . import q

KDB_INFO = "KDB+ %s %s" % tuple(q('.z.K,.z.k'))

style_dict = {
    # Prompt.
    Token.Generic.Prompt: '#884444',
    # Toolbar.
    Token.Toolbar: '#ffffff bg:#333333',
}


def get_bottom_toolbar_tokens(cli):
    """Return a list of tokens for the bottom toolbar"""
    mem = q('.Q.w', '') // 1024  # memory info in KiB
    return [(Token.Toolbar, "{0} {1.used}/{1.mphy} KiB".format(KDB_INFO, mem))]


history = InMemoryHistory()


def get_prompt_tokens(_):
    """Return a list of tokens for the prompt"""
    namespace = q(r'\d')
    if namespace == '.':
        namespace = ''
    return [(Token.Generic.Prompt, 'q%s)' % namespace)]


HSYM_RE = re.compile(r'.*`:([\w/.]*)$')


class QCompleter(Completer):
    """Completer for the q language"""

    def __init__(self):
        namespace = q(r'\d')
        res = q('.Q.res')
        dot_q = q('1_key .q')
        self.path_completer = PathCompleter()
        self.words_info = [(list(res), 'k'),
                           (list(dot_q), 'q'),
                           (list(q.key(namespace)), str(namespace))]

    def get_completions(self, document, complete_event):
        """Yield completions"""
        # Detect a file handle
        m = HSYM_RE.match(document.text_before_cursor)
        if m:
            text = m.group(1)
            doc = Document(text, len(text))
            for c in self.path_completer.get_completions(doc, complete_event):
                yield c
        else:
            # Get word/text before cursor.
            word_before_cursor = document.get_word_before_cursor(False)
            for words, meta in self.words_info:
                for a in words:
                    if a.startswith(word_before_cursor):
                        yield Completion(a, -len(word_before_cursor),
                                         display_meta=meta)


def cmdloop(self, intro=None):
    """A Cmd.cmdloop implementation"""
    style = style_from_pygments(BasicStyle, style_dict)
    self.preloop()
    stop = None
    while not stop:
        line = prompt(get_prompt_tokens=get_prompt_tokens, lexer=lexer,
                      get_bottom_toolbar_tokens=get_bottom_toolbar_tokens,
                      history=history, style=style, true_color=True,
                      on_exit='return-none', on_abort='return-none',
                      completer=QCompleter())
        if line is None or line.strip() == r'\\':
            raise SystemExit
        else:
            line = self.precmd(line)
            stop = self.onecmd(line)
        stop = self.postcmd(stop, line)
    self.postloop()



from __future__ import absolute_import

try:
    import numpy
    from numpy import ma
except ImportError:
    numpy = ma = None

import pytest

import pyq
from pyq import *
from pyq import Q_VERSION
from .test_k import K_INT_CODE, K_LONG_CODE

SYM_NA = int(K.int.na if Q_VERSION < 3.6 else K.long.na)
_PY3K = 1

pytestmark = pytest.mark.skipif(numpy is None, reason="numpy is not installed")
SIZE_OF_PTR = pyq._k.SIZEOF_VOID_P

SIMPLE_DTYPES = 'bool uint8 int16 int32 int64 float32 float64'.split()
# TODO: Add support for the "Y" unit.  What to do with ps and as?
# TIME_UNITS = 'Y M W D h m s ms us ns ps as'.split()
TIME_UNITS = 'M W D h m s ms us ns'.split()
DATETIME_DTYPES = ['datetime64[%s]' % unit for unit in TIME_UNITS]
TIMEDELTA_DTYPES = ['timedelta64[%s]' % unit for unit in TIME_UNITS if
                    unit != 'M']


@pytest.mark.parametrize(('t', 'r'), [
    ('?', 'b'),
    ('h', 'h'),
    ('i', 'i'),
    ('q', 'j'),
    ('f', 'e'),
    ('d', 'f'),
])
def test_ndarray(t, r):
    a = numpy.zeros(0, t)
    assert K(a) == k('"%c"$()' % r)

    a = numpy.zeros(1, t)
    assert K(a) == k(',0' + r)


def test_ma():
    x = ma.array([1.0, 2, 3])
    assert K(x) == k('1.0 2 3')

    x = ma.masked_values([1.0, 0, 2], 0)
    assert K(x) == k('1 0n 2')

    s = ma.masked_values(0.0, 0)
    assert K(s) == k('0n')


@pytest.mark.parametrize(('n', 'k'), [
    ('int8', '0x01'),
    ('int16', '1h'),
    ('int32', '1i'),
    ('int64', '1j'),
    ('float32', '1e'),
    ('float64', '1f'),
])
def test_scalar_conversion(n, k):
    x = getattr(numpy, n)(1)
    assert K(x) == q(k)


@pytest.mark.parametrize('a', [
    '0n', '1i', '1j', '0.1f', '1e', '1h',
    '010101b',
    '0xdeadbeef',
    '0 0n 1 0n',
    '0 0N 1 0N',
    '1 2 3 0Nh',
    '0.0 0.1 0Nf',
    '"abracadabra"',
    '"s p a c e d"',
    '0.0 0.1 0Ne',
    '1 2 3 0Nj',
    '1 2 3 0Ni',
])
def test_ma_array_roundtrip(a):
    x = q(a)
    m = ma.array(x)

    assert K(m) == x


def test_ma_conversion():
    x = q('1 0N 2')
    m = ma.array(x)

    assert isinstance(m.mask, numpy.ndarray)


@pytest.mark.parametrize(
    't,char,size', [
        (1, '?', 1),
        (4, 'B', 1),
        (5, 'h', 2),
        (6, K_INT_CODE, 4),
        (7, K_LONG_CODE, 8),
        (8, 'f', 4),
        (9, 'd', 8),
        (10, 'S', 1),
        # (11, 'O', SIZE_OF_PTR)  # issue #589
        (12, 'M', 8),
        (13, 'M', 8),
        (14, 'M', 8),
        # (15, 'M', 8),  # datetime
        (16, 'm', 8),
        (17, 'm', 8),
        (18, 'm', 8),
        (19, 'm', 8),
    ])
def test_empty(t, char, size):
    x = q('$[;()]', K._kh(t))
    a = numpy.array(x)

    assert a.dtype.char == char
    assert a.dtype.itemsize == size


def test_chartype():
    x = kp('abc')
    a = numpy.array(x)

    assert a.tolist() == [b'a', b'b', b'c']
    assert K(a) == x


def test_bytestype():
    x = q('0xabab')
    a = numpy.array(x)

    assert a.tolist() == [0xAB] * 2
    assert K(a) == x


@pytest.mark.xfail  # issue #591
def test_splayed_char(tmpdir):
    x = q('([]x:enlist"abc")')
    d = q('{sv[`;x,`x`]}', tmpdir)
    d.set(x)
    y = d.get.x
    a = numpy.array(y)

    assert y.inspect(b't') == 87
    assert a.dtype.char == 'S'
    assert a.tolist() == [b'a', b'b', b'c']


@pytest.mark.skipif("pyq.Q_VERSION < 3")
def test_symbol_list():
    x = K(['a'])
    a = numpy.array(x)
    assert numpy.array_equiv(a, ['a'])

    x = K(['a'] * 3)
    a = numpy.array(x, 'O')
    assert a[0] is a[1] is a[2] == 'a'


def test_settitem():
    a = numpy.array([1])
    a[0] = K(2)
    assert a[0] == 2


def test_issue652():
    a = numpy.array([(q('1h'), q('1i'), q('1j'))], dtype='h,i,q')
    assert a.dtype == numpy.dtype('h,i,q')


def test_datetime64_vector():
    a = numpy.array(['2001-01-01'] * 3, dtype='M8')
    assert K(a) == q('3#2001.01.01')

    a = numpy.array(['2001-01-01T00'], dtype='M8[ns]')
    assert K(a) == q('enlist 2001.01.01D00')

    a = numpy.array(['2001-01-01T00'], dtype='M8[h]')
    assert K(a) == q('enlist 2001.01.01D00')

    a = numpy.array(['2001-01-01T00'], dtype='M8[m]')
    assert K(a) == q('enlist 2001.01.01D00')

    a = numpy.array(['2001-01-01T00'], dtype='M8[s]')
    assert K(a) == q('enlist 2001.01.01D00')

    a = numpy.array(['2001-01-01T00'], dtype='M8[ms]')
    assert K(a) == q('enlist 2001.01.01D00')

    a = numpy.array(['2001-01-01T00'], dtype='M8[us]')
    assert K(a) == q('enlist 2001.01.01D00')


def test_datetime64_scalar():
    a = numpy.array('2001-01-01', dtype='M8')
    assert K(a) == q('2001.01.01')

    a = numpy.datetime64(1, 'Y')
    assert K(a) == q('1971.01m')

    a = numpy.datetime64(1, 'M')
    assert K(a) == q('1970.02m')

    a = numpy.datetime64(1, 'W')
    assert K(a) == q('1970.01.08')

    a = numpy.array('2001-01-01T00', dtype='M8[ns]')
    assert K(a) == q('2001.01.01D00')

    a = numpy.array('2001-01-01T00', dtype='M8[h]')
    assert K(a) == q('2001.01.01D00')

    a = numpy.array('2001-01-01T00', dtype='M8[m]')
    assert K(a) == q('2001.01.01D00')

    a = numpy.array('2001-01-01T00', dtype='M8[s]')
    assert K(a) == q('2001.01.01D00')

    a = numpy.array('2001-01-01T00', dtype='M8[ms]')
    assert K(a) == q('2001.01.01D00')

    a = numpy.array('2001-01-01T00', dtype='M8[us]')
    assert K(a) == q('2001.01.01D00')


def test_timedelta64_vector():
    a = numpy.array([1, 2, 3], dtype='m8[s]')
    assert K(a) == q('"n"$ 1 2 3 * 1000000000j')

    a = numpy.array([1, 2, 3], dtype='m8[ms]')
    assert K(a) == q('"n"$ 1 2 3 * 1000000j')

    a = numpy.array([1, 2, 3], dtype='m8[us]')
    assert K(a) == q('"n"$ 1 2 3 * 1000j')

    a = numpy.array([1, 2, 3], dtype='m8[ns]')
    assert K(a) == q('"n"$ 1 2 3j')


def test_timedelta64_scalar():
    a = numpy.timedelta64(1, 'W')
    assert K(a) == q('7D00:00:00')

    a = numpy.timedelta64(1, 'D')
    assert K(a) == q('1D00:00:00')

    a = numpy.timedelta64(1, 'h')
    assert K(a) == q('0D01:00:00')

    a = numpy.timedelta64(1, 'm')
    assert K(a) == q('0D00:01:00')

    a = numpy.timedelta64(1, 's')
    assert K(a) == q('0D00:00:01')

    a = numpy.timedelta64(1, 'ms')
    assert K(a) == q('0D00:00:00.001')

    a = numpy.timedelta64(1, 'us')
    assert K(a) == q('0D00:00:00.000001')

    a = numpy.timedelta64(1, 'ns')
    assert K(a) == q('0D00:00:00.000000001')


def test_record_arrays():
    a = numpy.array([(date(2001, 1, 1), 1, 3.14),
                     (date(2001, 1, 2), 2, 2.78), ],
                    dtype=[('date', 'M8[D]'), ('num', 'i2'), ('val', 'f')])
    t = K(a)
    assert t.date == q('2001.01.01 2001.01.02')
    assert t.num == q('1 2h')
    assert t.val == q('3.14 2.78e')

    r = K(a[0])
    assert r == q("`date`num`val!(2001.01.01;1h;3.14e)")


def test_unsupported_dtype_errors():
    with pytest.raises(TypeError):
        a = numpy.array('abc', dtype='S3')
        K(a)


@pytest.mark.parametrize('e', ['sym', 'other'])
def test_enum_conversion(q, e):
    x = q('`%s?`a`b`c' % e)
    a = numpy.asarray(x.data)
    a.tolist() == [0, 1, 2]


@pytest.mark.parametrize('data, dtype, kstr', [
    ([1, 0, 1], bool, '101b'),
    ([1, 2, 3], 'h', '1 2 3h'),
    ([1, 2, 3], 'i', '1 2 3i'),
    ([1, 2, 3], 'q', '1 2 3j'),
    ([1, 2, 3], 'f', '1 2 3e'),
    ([1, 2, 3], 'd', '1 2 3f'),
])
def test_from_array_struct(data, dtype, kstr):
    a = numpy.array(data, dtype)
    x = K._from_array_struct(a.__array_struct__)
    assert x == q(kstr)


def test_from_array_struct_errors():
    with pytest.raises(TypeError):
        K._from_array_struct()
    # XXX: Shouldn't this be a TypeError?
    with pytest.raises(ValueError):
        K._from_array_struct(None)
    # XXX: Shouldn't this be a NotImplementedError?
    # a = numpy.zeros((1, 1))
    # with pytest.raises(ValueError):
    #     K._from_array_struct(a.__array_struct__)
    a = numpy.array(['', None], 'O')
    # XXX: Shouldn't this be a TypeError?
    with pytest.raises(ValueError):
        K._from_array_struct(a.__array_struct__)


def test_numpy_in_versions(capsys):
    versions()
    out = capsys.readouterr()[not _PY3K]
    assert 'NumPy' in out


def test_no_dtype():
    class NoDtype(numpy.ndarray):
        """A helper class to test error branches"""

        @property
        def dtype(self):
            raise AttributeError('intentional error')

    x = NoDtype([0], dtype='M8[D]')
    with pytest.raises(AttributeError) as info:
        K(x)
    assert 'intentional error' in str(info.value)


@pytest.mark.parametrize('kstr, typestr', zip(
    ('0b;0x00;0h;0i;0j;0e;0.0;" ";`;2000.01m;2000.01.01;'
     '2000.01.01T00:00:00.000;00:00;00:00:00;00:00:00.000').split(';'),
    ('<b1,<u1,<i2,<i4,<i8,<f4,<f8,<S1,'
     '|O%d,<i4,<i4,<f8,<i4,<i4,<i4' % SIZE_OF_PTR).split(',')))
def test_array_typestr(kstr, typestr):
    x = q(kstr)
    assert x.__array_typestr__ == typestr


def test_there_and_back_again(q):
    x = q.til(10)
    a = numpy.asarray(x)
    y = K(a)
    assert x is y


def test_table_to_array():
    t = q('([]a:1 2;b:"xy")')
    a = numpy.array(t)
    assert a.tolist() == [(1, b'x'),
                          (2, b'y')]


@pytest.mark.parametrize('dtype',
                         SIMPLE_DTYPES + TIMEDELTA_DTYPES + DATETIME_DTYPES)
def test_table_to_array_all_types(dtype):
    c = numpy.zeros(1, dtype)
    t = +q('!', ['c'], (c,))
    a = numpy.asarray(t)
    assert a.dtype.names == ('c',)
    assert numpy.array_equiv(c, a['c'])


@pytest.mark.skipif("q('.z.K') < 3")
@pytest.mark.parametrize('dtype', SIMPLE_DTYPES)
def test_2d(dtype):
    a = numpy.zeros((2, 3), dtype)
    x = K(a)
    y = K(a.flatten())
    assert x.raze == y


@pytest.mark.skipif("q('.z.K') < 3.4")
@pytest.mark.parametrize('dtype', SIMPLE_DTYPES)
def test_3d(dtype):
    a = numpy.zeros((2, 3, 2), dtype)
    x = K(a)
    y = K(a.flatten())
    assert x.raze.raze == y


@pytest.mark.parametrize('dtype', SIMPLE_DTYPES)
def test_0d(dtype):
    a = numpy.ones((), dtype)
    x = K(a)
    assert x.enlist[0] == 1


def test_enum_to_array(q):
    x = q('`sym?`a`b')
    a = numpy.array(['a', 'b'], dtype=object)
    assert numpy.array_equiv(x, a)


@pytest.mark.skipif("q('.z.K') < 3.4")
@pytest.mark.parametrize('dtype', SIMPLE_DTYPES)
def test_2d_roundtrip(dtype):
    a = numpy.zeros((3, 2), dtype)
    x = K(a)
    b = numpy.array(x)
    assert numpy.array_equiv(a, b)


def test_priority():
    a = numpy.arange(2)
    b = q.til(2)
    assert isinstance(a + b, K)
    assert isinstance(b + a, K)


class A(object):
    @property
    def __array_struct__(self):
        raise TypeError


def test_broken_array_struct():
    with pytest.raises(TypeError):
        K(A())


def test_broken_mask():
    class B(object):
        a = numpy.zeros(0)

        @property
        def __array_struct__(self):
            return self.a.__array_struct__

        @property
        def mask(self):
            raise TypeError

    with pytest.raises(TypeError):
        K(B())


def test_strided_nd():
    a = numpy.zeros((2, 2, 2))
    with pytest.raises(ValueError):
        K(a.diagonal())


def test_0d_str():
    a = numpy.array('x', 'O')
    assert K(a) == 'x'


def test_call_python(q):
    def f():
        return numpy.array([1.0, 2.0])
    q.f = f
    assert q("f()") == [1.0, 2.0]


def test_enum_to_numpy(q):
    x = q('`sym?`a`b`')
    assert numpy.asarray(x.data).tolist() == [0, 1, SYM_NA]



from __future__ import absolute_import
import os
import sys
import subprocess
import platform
import pytest

WIN = platform.system() == "Windows"

TEST_P = """\
import sys
from pyq import q
def test():
    pass
assert len(sys.argv) == len(q(".z.x")) + 1, "test argv"
q("p)print('ok')")
sys.exit(0)
"""


def test_test_p0(tmpdir):
    test_p = tmpdir.join('test.p')
    test_p.write(TEST_P)
    with open(os.devnull) as null:
        out = subprocess.check_output([os.environ['QBIN'], str(test_p)],
                                      stderr=subprocess.PIPE, stdin=null)
    assert b'ok' in out


def test_test_p_multiple_arugments(tmpdir):
    p = tmpdir.join('test.p')
    p.write("""\
import sys
from pyq import q

def test(a, b, c):
    return a + b + c

q("p)print('ok' if test(1, 2, 3) else 'fail')")
sys.exit(0)
""")
    with open(os.devnull) as null:
        out = subprocess.check_output([os.environ['QBIN'], str(p)],
                                      stderr=subprocess.PIPE, stdin=null)
    assert b'ok' in out


@pytest.mark.skipif(WIN, reason="This test hangs on Windows.")
def test_test_p_exception(tmpdir):
    p = tmpdir.join('test.q')
    p.write("p)1+'a'")
    p = subprocess.Popen([os.environ['QBIN'], str(p)], stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    assert out == b''
    traceback = b'''\
Traceback (most recent call last):
  File "<string>", line 1, in <module>
TypeError: unsupported operand type(s) for +: 'int' and 'str'
'''
    assert traceback in err


def test_p__file__qbin(tmpdir):
    p = tmpdir.join('test.p')
    p.write("import sys\nprint(__file__)\nsys.exit(0)")
    with open(os.devnull) as null:
        out = subprocess.check_output([os.environ['QBIN'], str(p)],
                                      stderr=subprocess.PIPE, stdin=null)
    assert out.strip().endswith(str(p).encode())



"""pyq - python for kdb+"""
import itertools
from datetime import datetime, date, time
from collections import Mapping as _Mapping

import builtins
import sys
import os

try:
    import numpy as _np
except ImportError:
    _np = None

try:
    from ._k import K as _K, error as kerr, Q_VERSION, Q_DATE, Q_OS
except ImportError:
    import ctypes
    import platform

    if not hasattr(ctypes.CDLL(None), 'b9'):
        message = ("Importing the pyq package from "
                   "standalone python is not supported. ")
        if platform.system() == 'Windows':
            message += "Run path\\to\\q.exe python.q."
        else:
            message += "Use pyq executable."
        raise ImportError(message)
    raise

try:
    from .version import version as __version__
except ImportError:
    __version__ = 'unknown'

__metaclass__ = type

# Convenience constant to select code branches according to Q_VERSION
_KX3 = Q_VERSION >= 3

# List of q builtin functions that are not defined in .q.
# NB: This is similar to .Q.res, but excludes non-function constructs
# such as "do", "if", "select" etc. We also exclude the "exit" function
# because it is rarely safe to use it to exit from pyq.
_Q_RES = ['abs', 'acos', 'asin', 'atan', 'avg', 'bin',  'cor', 'cos',
          'cov', 'dev', 'div', 'ema', 'enlist', 'exp', 'getenv', 'in',
          'insert', 'last', 'like', 'log', 'max', 'min', 'prd', 'reval',
          'setenv', 'sin', 'sqrt', 'ss', 'sum',
          'tan', 'var', 'wavg', 'within', 'wsum', 'xexp']

if _KX3 and Q_DATE >= date(2012, 7, 26):
    # binr was introduced in kdb+3.0 2012.07.26
    _Q_RES.append('binr')

if Q_VERSION >= 3.6:
    _Q_RES.append('hopen')


class K(_K):
    """proxies for kdb+ objects

    >>> q('2005.01.01 2005.12.04')
    k('2005.01.01 2005.12.04')

    Iteration over simple lists produces python objects

    >>> list(q("`a`b`c`d"))
    ['a', 'b', 'c', 'd']

    Iteration over q tables produces q dictionaries

    >>> list(q("([]a:`x`y`z;b:1 2 3)"))
    [k('`a`b!(`x;1)'), k('`a`b!(`y;2)'), k('`a`b!(`z;3)')]

    Iteration over a q dictionary iterates over its key

    >>> list(q('`a`b!1 2'))
    ['a', 'b']

    as a consequence, iteration over a keyed table is the same as
    iteration over its key table

    >>> list(q("([a:`x`y`z]b:1 2 3)"))
    [k('(,`a)!,`x'), k('(,`a)!,`y'), k('(,`a)!,`z')]

    Callbacks into python

    >>> def f(x, y):
    ...     return x + y
    >>> q('{[f]f(1;2)}', f)
    k('3')

    Buffer protocol

    The following session illustrates how buffer protocol implemented by
    K objects can be used to write data from Python streams directly yo kdb+.

    Create a list of chars in kdb+

    >>> x = kp('xxxxxx')

    Open a pair of file descriptors

    >>> r, w = os.pipe()

    Write 6 bytes to the write end

    >>> os.write(w, b'abcdef')
    6

    Read from the read-end into x

    >>> f = os.fdopen(r, mode='rb')
    >>> f.readinto(x)
    6

    Now x contains the bytes that were sent through the pipe

    >>> x
    k('"abcdef"')

    Close the descriptors and the stream

    >>> os.close(w); f.close()

    Low level interface

    The K type provides a set of low level functions that are similar
    to the C API provided by the `k.h header <http://kx.com/q/c/c/k.h>`_.
    The C API functions that return K objects in C are implemented as
    class methods that return instances of K type.

    Atoms

    >>> K._kb(True), K._kg(5), K._kh(42), K._ki(-3), K._kj(2**40)
    (k('1b'), k('0x05'), k('42h'), k('-3i'), k('1099511627776'))

    >>> K._ke(3.5), K._kf(1.0), K._kc(b'x'), K._ks('xyz')
    (k('3.5e'), k('1f'), k('"x"'), k('`xyz'))

    >>> K._kd(0), K._kz(0.0), K._kt(0)
    (k('2000.01.01'), k('2000.01.01T00:00:00.000'), k('00:00:00.000'))


    Tables and dictionaries

    >>> x = K._xD(k('`a`b`c'), k('1 2 3')); x, K._xT(x)
    (k('`a`b`c!1 2 3'), k('+`a`b`c!1 2 3'))

    Keyed table

    >>> t = K._xD(K._xT(K._xD(k(",`a"), k(",1 2 3"))),
    ...           K._xT(K._xD(k(",`b"), k(",10 20 30"))))
    >>> K._ktd(t)
    k('+`a`b!(1 2 3;10 20 30)')

    """
    # Lighten the K objects by preventing the automatic creation of
    # __dict__ and __weakref__ for each instance.
    __slots__ = ()

    # Helper methods for use in C implementation of __new__

    def _set_mask(self, mask):
        return q("{?[y;((),x)0N;x]}", self, mask)

    @classmethod
    def _from_record_array(cls, x):
        fields = [f for f, t in x.dtype.descr]
        k = q('!', list(fields), [K(x[f]) for f in fields])
        if x.ndim:
            k = k.flip
        return k

    @classmethod
    def _from_sequence(cls, x, elm=None):
        r = cls._ktn(0, 0)
        g = iter(x)
        try:
            i = next(g)
        except StopIteration:
            return r
        en = _K_k(0, 'enlist')
        r._ja(en)
        if elm is None:
            elm = cls
        for i in itertools.chain([i], g):
            i = elm(i)
            # Symbols and lists require special treatment
            if i._t in (-11, 11, 0):
                i = i.enlist
            r._ja(i)
        return r.eval

    @classmethod
    def _convert(cls, x):
        for t in type(x).mro():
            c = converters.get(t)
            if c is not None:
                return c(x)
        return cls._from_sequence(x)

    def __reduce_ex__(self, proto):
        x = self._b9(1, self)
        b = memoryview(x).tobytes()
        return (d9, (b,))

    def __getitem__(self, x):
        """
        >>> k("10 20 30 40 50")[k("1 3")]
        k('20 40')
        >>> k("`a`b`c!1 2 3")['b']
        2
        """
        try:
            return _K.__getitem__(self, x)
        except (TypeError, NotImplementedError):
            pass
        try:
            start, stop, step = x.indices(len(self))
        except AttributeError:
            i = K(x)
            if self._t == 99 and i._t < 0:
                return self.value[self._k(0, "?", self.key, i)]
            else:
                return self._k(0, "@", self, i)

        if step == 1:
            return self._k(0, "sublist", self._J([start, stop - start]), self)
        # NB: .indices() cannot return step=0.
        i = start + step * q.til(max(0, (stop - start) // step))

        return self._k(0, "{$[99=type x;(key[x]y)!value[x]y;x y]}", self, i)

    def __getattr__(self, a):
        """table columns can be accessed via dot notation

        >>> q("([]a:1 2 3; b:10 20 30)").a
        k('1 2 3')
        """
        t = self._t
        if t == 98 and not a.startswith('_'):
            return self._k(0, '{x`%s}' % a, self)

        if t == 99:
            if self._k(0, "{11h~type key x}", self):
                if a == 'items':
                    # NB: Workaround for a bug in OrderedDict in Python 3.5.
                    # See http://bugs.python.org/issue27576 for details.
                    raise AttributeError
                return self._k(0, '{x`%s}' % a, self)

            return self._k(0, '{(0!x)`%s}' % a, self)

        if 12 <= abs(t) < 20:
            try:
                return self._k(0, "`%s$" % a, self)
            except kerr:
                pass

        raise AttributeError(a)

    _fields = " g@ ghijefgsjiifjiii"

    def __int__(self):
        """converts K scalars to python int

        >>> [int(q(x)) for x in '1b 2h 3 4e `5 6.0 2000.01.08'.split()]
        [1, 2, 3, 4, 5, 6, 7]
        """
        t = self._t
        if t >= 0:
            raise TypeError("cannot convert non-scalar to int")
        return int(self.inspect(self._fields[-t]))

    def __float__(self):
        """converts K scalars to python float

        >>> [float(q(x)) for x in '1b 2h 3 4e `5 6.0 2000.01.08'.split()]
        [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
        """
        t = self._t
        if t >= 0:
            raise TypeError("cannot convert non-scalar to float")
        return float(self.inspect(self._fields[-t]))

    def __index__(self):
        t = self._t
        if -5 >= t >= -7:
            return int(self)
        raise TypeError("Only scalar short/int/long K objects "
                        "can be converted to an index")

    def __bytes__(self):
        t = self._t
        if -5 >= t >= -7:
            return bytes(int(self))
        if 0 < abs(t) < 11:
            if abs(t) == 2:
                # A work-around while .data is not implemented for guid type
                from uuid import UUID
                x = q('(),', self)  # ensure that x is a list
                return b''.join(UUID(int=i).bytes for i in x)
            return bytes(self.data)
        raise BufferError("k object of type %d" % t)

    def __eq__(self, other):
        """
        >>> K(1) == K(1)
        True
        >>> K(1) == None
        False
        """
        try:
            other = K(other)
        except TypeError:
            return False
        return bool(k('~')(self, other))

    def __ne__(self, other):
        """
        >>> K(1) != K(2)
        True
        """
        return bool(k('~~')(self, other))

    def __contains__(self, item):
        """membership test

        >>> 1 in q('1 2 3')
        True

        >>> 'abc' not in q('(1;2.0;`abc)')
        False
        """
        if self._t:
            x = q('in', item, self)
        else:
            x = q('{sum x~/:y}', item, self)
        return bool(x)

    def keys(self):
        """returns q('key', self)

        Among other uses, enables interoperability between q and
        python dicts.

        >>> from collections import OrderedDict
        >>> OrderedDict(q('`a`b!1 2'))
        OrderedDict([('a', 1), ('b', 2)])
        >>> d = {}; d.update(q('`a`b!1 2'))
        >>> list(sorted(d.items()))
        [('a', 1), ('b', 2)]
        """
        return self._k(0, 'key', self)

    def show(self, start=0, geometry=None, output=None):
        """pretty-print data to the console

        (similar to q.show, but uses python stdout by default)

        >>> x = q('([k:`x`y`z]a:1 2 3;b:10 20 30)')
        >>> x.show()  # doctest: +NORMALIZE_WHITESPACE
        k| a b
        -| ----
        x| 1 10
        y| 2 20
        z| 3 30

        the first optional argument, 'start' specifies the first row to be
        printed (negative means from the end)

        >>> x.show(2)  # doctest: +NORMALIZE_WHITESPACE
        k| a b
        -| ----
        z| 3 30

        >>> x.show(-2)  # doctest: +NORMALIZE_WHITESPACE
        k| a b
        -| ----
        y| 2 20
        z| 3 30

        the geometry is the height and width of the console

        >>> x.show(geometry=[4, 6])
        k| a..
        -| -..
        x| 1..
        ..

        """
        if output is None:
            output = sys.stdout

        if geometry is None:
            geometry = q.value(kp("\\c"))
        else:
            geometry = self._I(geometry)

        if start < 0:
            start += q.count(self)

        # Make sure nil is not passed to a q function
        if self._id() != nil._id():
            r = self._show(geometry, start)
        else:
            r = '::\n'

        if isinstance(output, type):
            return output(r)
        try:
            output.write(r)
        except TypeError:
            output.write(str(r))

    # See issue #665
    def decode(self, encoding='utf-8', errors='strict'):
        return bytes(self).decode(encoding, errors)

    def _seu(self, what, columns, by, where, kwds):
        args = [self]
        anames = ['self']
        if kwds:
            extra = sorted(kwds.keys())
            args.extend(kwds[name] for name in extra)
            anames.extend(extra)

        if not isinstance(columns, str):
            columns = ','.join(str(x) for x in columns)
        query = "{[%s]%s %s " % (';'.join(anames), what, columns)
        if by:
            if not isinstance(by, str):
                by = ','.join(str(x) for x in by)
            query += " by " + by
        query += " from self"
        if where:
            if not isinstance(where, str):
                where = ','.join(str(x) for x in where)
            query += " where " + where
        query += '}'

        return q(query, *args)

    def select(self, columns=(), by=(), where=(), **kwds):
        """select from self

        >>> t = q('([]a:1 2 3; b:10 20 30)')
        >>> t.select('a', where='b > 20').show()
        a
        -
        3
        """
        return self._seu('select', columns, by, where, kwds)

    def exec(self, columns=(), by=(), where=(), **kwds):
        """exec from self

        >>> t = q('([]a:1 2 3; b:10 20 30)')
        >>> t.exec('a', where='b > 10').show()
        2 3
        """
        return self._seu('exec', columns, by, where, kwds)

    exec_ = exec  # TODO: deprecate

    def update(self, columns=(), by=(), where=(), **kwds):
        """update from self

        >>> t = q('([]a:1 2 3; b:10 20 30)')
        >>> t.update('a*2',
        ...          where='b > 20').show()  # doctest: +NORMALIZE_WHITESPACE
        a b
        ----
        1 10
        2 20
        6 30
        """
        return self._seu('update', columns, by, where, kwds)

    def delete(self, columns=(), where=()):
        """delete from self

        >>> t = q('([]a:1 2 3; b:10 20 30)')
        >>> t.delete('a').show()  # doctest: +NORMALIZE_WHITESPACE
        b
        --
        10
        20
        30
        """
        return self._seu('delete', columns, (), where, {})

    @property
    def ss(self):
        if self._t == 10:
            return q.ss(self)
        return q('`ss$', self)

    if _np is not None:
        @property
        def _mask(self):
            x = self._get_null()
            if x is None:
                return _np.ma.nomask
            else:
                return _np.asarray(x)

        from ._n import array as __array__

        __array_priority__ = 20

    __doc__ += """
    K objects can be used in Python arithmetic expressions

    >>> x, y, z = map(K, (1, 2, 3))
    >>> print(x + y, x * y,
    ...       z/y, x|y, x&y, abs(-z))  #doctest: +NORMALIZE_WHITESPACE
    3 2 1.5 2 1 3

    Mixing K objects with python numbers is allowed

    >>> 1/q('1 2 4')
    k('1 0.5 0.25')
    >>> q.til(5)**2
    k('0 1 4 9 16f')
    """

    def __format__(self, fmt):
        if fmt:
            return format(self._pys(), fmt)
        return str(self)

    def __sizeof__(self):
        return object.__sizeof__(self) + int(self._sizeof())

    def __fspath__(self):
        """Return the file system path representation of the object."""
        if self._t != -11:  # symbol
            raise TypeError
        sym = str(self)
        if not sym.startswith(':'):
            raise TypeError
        return sym[1:]

    def __complex__(self):
        """Called to implement the built-in function complex()."""
        if self._t != 99 or self.key != ['re', 'im']:
            return complex(float(self))
        return complex(float(self.re), float(self.im))

    @classmethod
    def dict(cls, *args, **kwds):
        """Construct a q dictionary

        K.dict() -> new empty q dictionary (q('()!()')
        K.dict(mapping) -> new dictionary initialized from a mapping object's
            (key, value) pairs
        K.dict(iterable) -> new dictionary initialized from an iterable
            yielding (key, value) pairs
        K.dict(**kwargs) -> new dictionary initialized with the name=value
            pairs in the keyword argument list.
            For example: K.dict(one=1, two=2)
        """
        if args:
            if len(args) > 1:
                raise TypeError("Too many positional arguments")
            x = args[0]
            keys = []
            vals = []
            try:
                x_keys = x.keys
            except AttributeError:
                for k, v in x:
                    keys.append(k)
                    vals.append(v)
            else:
                keys = x_keys()
                vals = [x[k] for k in keys]
            return q('!', keys, vals)
        else:
            if kwds:
                keys = []
                vals = []
                for k, v in kwds.items():
                    keys.append(k)
                    vals.append(v)
                return q('!', keys, vals)
            else:
                return q('()!()')

    @classmethod
    def table(cls, *args, **kwds):
        if args or kwds:
            return cls.dict(*args, **kwds).flip
        else:
            raise TypeError("A table must have at least one column")


def _q_builtins():
    from keyword import iskeyword

    # Allow _q_builtins() to be called before q is defined
    def q(x):
        return K._k(0, x)

    kver = q('.z.K').inspect('f')
    names = _Q_RES + list(q("1_key[.q]except`each`over`scan"))
    if kver < 3.4:
        # ema is present since kdb+3.4
        q(r'.q.ema:{first[y]("f"$1-x)\x*y}')
    if kver < 3.3:
        # restricted eval was added in 3.3
        q(r'.q.reval:eval')

    pairs = []
    for x in names:
        attr = x
        if iskeyword(x):
            attr += '_'
        pairs.append((attr, q(x)))

    return pairs


def _genmethods(cls, obj):
    q(r'\l pyq-operators.q')
    cls._show = q('{` sv .Q.S[y;z;x]}')
    cls._sizeof = q('.pyq.sizeof')
    for spec, verb in [
        ('add', '+'), ('sub', '-'), ('rsub', '{y-x}'),
        ('mul', '*'), ('pow', 'xexp'), ('rpow', '{y xexp x}'),
        ('xor', '^'), ('rxor', '{y^x}'),
        ('truediv', '%'), ('rtruediv', '{y%x}'),
        ('floordiv', 'div'), ('rfloordiv', '{y div x}'),
        ('and', '&'), ('or', '|'),
        ('mod', 'mod'), ('rmod', '{y mod x}'), ('invert', 'not'),
        ('pos', '{@[flip;x;x]}'), ('neg', '-:'), ('abs', 'abs'),
        # x @ y - composition if y is a function, x[y] otherwise.
        ('matmul', "{$[100>type y;x y;'[x;y]]}"),
        ('rmatmul', "{$[100>type x;y x;'[y;x]]}"),
    ]:
        setattr(cls, '__%s__' % spec, q(verb))

    for spec in 'add mul and or'.split():
        setattr(cls, '__r%s__' % spec, getattr(cls, '__%s__' % spec))

    for x, f in _q_builtins():
        if not hasattr(cls, x):
            setattr(cls, x, f)
            obj.__dict__[x] = f

    def cmp_op(op):
        def dunder(self, other):
            other = K(other)
            if self._t < 0 and other._t < 0:
                return bool(q(op, self, other))
            else:
                raise NotImplementedError

        return dunder

    for spec, verb in [('gt', '>'), ('lt', '<'), ('ge', '>='), ('le', '<=')]:
        setattr(cls, '__%s__' % spec, cmp_op(verb))

    # Shift operators
    for x in ['', 'r']:
        for y in 'lr':
            op = x + y + 'shift'
            setattr(cls, '__%s__' % op, q('.pyq.' + op))


def d9(x):
    """like K._d9, but takes python bytes"""
    return K._d9(K._kp(x))


def k(m, *args):
    return K._k(0, 'k)' + m, *map(K, args))


class _Q(object):
    """a portal to kdb+"""

    def __init__(self):
        object.__setattr__(self, '_cmd', None)
        object.__setattr__(self, '_q_names',
                           [name for name, _ in _q_builtins()])

    def __call__(self, m=None, *args):
        """Execute q code."""
        try:
            return K._k(0, m, *map(K, args))
        except TypeError:
            if m is not None:
                raise
            if self._cmd is None:
                from .cmd import Cmd
                object.__setattr__(self, '_cmd', Cmd())
            self._cmd.cmdloop()

    def __getattr__(self, attr, _k=K._k):
        try:
            return _k(0, attr.rstrip('_'))
        except kerr:
            pass
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        self("@[`.;;:;]", attr, value)

    def __delattr__(self, attr):
        k = K._k
        k(0, "delete %s from `." % attr)

    def __dir__(self):
        return self._q_names + list(self.key('.'))


q = _Q()
nil = q('(value +[;0])1')

show = K.show


def _ni(x):
    r = q('()')
    append = q(',')
    for i in x:
        r = append(r, K(i))
    return r


_X = {K: K._K, str: K._S, int: (K._J if _KX3 else K._I),
      float: K._F, date: K._D, time: _ni, datetime: _ni, bool: K._B}


def _listtok(x):
    if x:
        for i in x:
            if i is not None:
                break
        c = _X.get(type(i))
        if c is not None:
            try:
                return c(x)
            except (TypeError, ValueError):
                pass
        return K._from_sequence(x)
    return K._ktn(0, 0)


_X[list] = lambda x: K([K(i) for i in x])


def _tupletok(x):
    try:
        fields = x._fields
    except AttributeError:
        return K._from_sequence(x)
    else:
        return K._xD(K(fields), K._from_sequence(x))


kp = K._kp

converters = {
    list: _listtok,
    tuple: _tupletok,
    type(lambda: 0): K._func,
    type(sum): K._func,
    dict: lambda x: K._xD(K(x.keys()), K(x.values())),
    complex: lambda z: K._xD(K._S(['re', 'im']), K._F([z.real, z.imag])),
    bytes: K._kp,
    bytearray: K._from_memoryview,
    memoryview: K._from_memoryview,
}


###############################################################################
# Lazy addition of converters
###############################################################################
lazy_converters = {
    'uuid': [('UUID', lambda u: K._kguid(u.int))],
    'py._path.local': [
        ('LocalPath',
            (lambda p: K(':' + p.strpath)) if os.sep == '/' else
            lambda p: K(':' + p.strpath.replace(os.sep, '/')))
    ],
    'pathlib': [('PurePath', lambda p: K(':' + p.as_posix()))],
}

lazy_converters['pathlib2'] = lazy_converters['pathlib']
lazy_converters['array'] = [('array', K._from_memoryview)]


# If module is already loaded, register converters for its classes
# right away.
def _pre_register_converters():
    for name, pairs in lazy_converters.items():
        mod = sys.modules.get(name)
        if mod is not None:
            for cname, conv in pairs:
                converters[getattr(mod, cname)] = conv


_pre_register_converters()
del _pre_register_converters

# Replace builtin import to add lazy registration logic
_imp = builtins.__import__


def __import__(name, globals={}, locals={}, fromlist=[], level=0,
               _imp=_imp, _c=converters, _lc=lazy_converters):
    m = _imp(name, globals, locals, fromlist, level)
    pairs = _lc.get(m.__name__)
    if pairs is not None:
        _c.update((getattr(m, cname), conv) for cname, conv in pairs)
    return m


builtins.__import__ = __import__
###############################################################################

_genmethods(K, q)
del _genmethods, _imp


def versions():
    """Report versions"""
    stream = sys.stdout
    print('PyQ', __version__, file=stream)
    if _np is not None:
        print('NumPy', _np.__version__, file=stream)
    print('KDB+ %s (%s) %s' % tuple(q('.z.K,.z.k,.z.o')), file=stream)
    print('Python', sys.version, file=stream)


###############################################################################
# Casts and constructors
###############################################################################
def _gendescriptors(cls, string_types=(type(b''), type(u''))):
    cls._Z = NotImplemented
    if Q_VERSION < 3:
        cls._UU = cls._kguid = NotImplemented
    types = [
        # code, char, name, vector, scalar
        (1, 'b', 'boolean', cls._B, cls._kb),
        (2, 'g', 'guid', cls._UU, cls._kguid),
        (4, 'x', 'byte', cls._G, cls._kg),
        (5, 'h', 'short', cls._H, cls._kh),
        (6, 'i', 'int', cls._I, cls._ki),
        (7, 'j', 'long', cls._J, cls._kj),
        (8, 'e', 'real', cls._E, cls._ke),
        (9, 'f', 'float', cls._F, cls._kf),
        (11, 's', 'symbol', cls._S, cls._ks),
        (12, 'p', 'timestamp', cls._P, cls._kpz),
        (13, 'm', 'month', cls._M, cls._km),
        (14, 'd', 'date', cls._D, cls._kd),
        (15, 'z', 'datetime', cls._Z, cls._kz),
        (16, 'n', 'timespan', cls._N, cls._knz),
        (17, 'u', 'minute', cls._U, cls._ku),
        (18, 'v', 'second', cls._V, cls._kv),
        (19, 't', 'time', cls._T, cls._kt),
    ]

    class Desc:
        def __init__(self, code, char, name, vector, scalar):
            self.code = cls._kh(code)
            self.char = char
            self.name = name
            self.vector = vector
            self.scalar = scalar

        def make_constructor(self):
            def constructor(x):
                # If x is already K - check the type and either
                # pass it through or cast to the type needed.
                if isinstance(x, K):
                    if x.type.abs == self.code:
                        return x
                    else:
                        return cls._k(0, '$', self.code, x)
                if isinstance(x, _Mapping):
                    return cls._xD(cls(x.keys()), constructor(x.values()))
                try:
                    return self.vector(x)
                except TypeError:
                    pass
                try:
                    return self.scalar(x)
                except TypeError:
                    return cls._from_sequence(x, constructor)

            constructor.__name__ = 'K.' + self.name
            if self.code > 4 and int(self.code) != 11:
                constructor.inf = q('0W' + self.char)
                constructor.na = q('0N' + self.char)
            elif self.name == 'guid':
                constructor.na = q('0Ng')

            return constructor

        def __get__(self, instance, owner):
            if instance is None:
                return self.make_constructor()
            # Make sure dict keys and table columns have priority over casts
            name = self.name
            if instance._t == 98 and name in instance.cols:
                return instance[name]
            if instance._t == 99:  # keyed table or dict
                key = instance.key
                if key._t == 11 and name in key:
                    return instance[name]
                if key._t == 98 and name in instance.cols:
                    return instance.exec_(name)

            return cls._k(0, '$', self.code, instance)

    for code, char, name, vector, scalar in types:
        setattr(cls, name, Desc(code, char, name, vector, scalar))

    # Special case: string
    def make_strings(x):
        if isinstance(x, string_types):
            return cls._kp(x)
        if isinstance(x, _Mapping):
            return cls._xD(cls(x.keys()), make_strings(x.values()))
        return cls._from_sequence(x, make_strings)

    class StringDesc:
        def __get__(self, instance, owner):
            if instance is None:
                return make_strings
            # NB: As a reserved word, "string" cannot be a column name but
            # can be a key in a dictionary
            if instance._t == 99:
                key = instance.key
                if key._t == 11 and 'string' in key:
                    return instance['string']

            return cls._k(0, 'string', instance)

    cls.string = StringDesc()

    # Special case: char (like string, but may return a scalar char.)
    def make_chars(x):
        if isinstance(x, str):
            x = x.encode('utf8')
        if isinstance(x, bytes):
            if len(x) == 1:
                return cls._kc(x)
            else:
                return cls._kp(x)
        if isinstance(x, _Mapping):
            return cls._xD(cls(x.keys()), make_chars(x.values()))
        if not x:
            return cls._kp('')
        return cls._from_sequence(x, make_chars)

    make_chars.inf = q('0Wc')
    make_chars.na = q('0Nc')

    class CharDesc:
        def __get__(self, instance, owner):
            if instance is None:
                return make_chars
            # Make sure dict keys and table columns have priority over casts
            name = 'char'
            if instance._t == 98 and name in instance.cols:
                return instance[name]
            if instance._t == 99:  # keyed table or dict
                key = instance.key
                if key._t == 11 and name in key:
                    return instance[name]
                if key._t == 98 and name in instance.cols:
                    return instance.exec_(name)

            return cls._k(0, '`char$', instance)

    cls.char = CharDesc()


_gendescriptors(K)
del _gendescriptors


def _genadverbs(cls):
    adverbs = [
        'each',   # '
        'over',   # /
        'scan',   # \
        'prior',  # ':
        'sv',     # /: - each-right
        'vs',     # \: - each-left
     ]
    for i, a in enumerate(adverbs):
        x = cls._ktj(103, i)
        setattr(cls, a, x)


_genadverbs(K)
del _genadverbs


# Traceback support
_K_k = K._k
_K_call = K.__call__


def _set_excepthook(origexcepthook):
    def excepthook(exctype, value, traceback):
        origexcepthook(exctype, value, traceback)
        a = value.args
        if exctype is kerr and len(a) > 1:
            sbt = _K_k(0, '.Q.sbt', a[1])
            print("kdb+ backtrace:\n%s" % sbt,
                  file=sys.stderr, end='')

    sys.excepthook = excepthook


_val = q.value
_enl = q.enlist


def _trp_k(cls_, h, m, *args):
    if h != 0:
        return cls_._k(h, m, args)
    f = cls_._trp(_val, K._knk(1, kp(m)))
    if args:
        return cls_._trp(f, K(args))
    else:
        return f


def _trp_call(*args, **kwds):
    f = args[0]
    args = args[1:]
    if f.type < 100:
        return _K_call(f, *args, **kwds)
    if kwds:
        args = f._callargs(*args, **kwds)
    if not args:
        args = [None]
    args = K._knk(len(args), *map(K, args))
    return f._trp(args)


if 'PYQ_BACKTRACE' in os.environ and q('.z.K >= 3.5'):
    _set_excepthook(sys.excepthook)
    K._k = classmethod(_trp_k)
    K.__call__ = _trp_call



"""q REPL"""
import cmd as _cmd

from . import q, kerr, Q_OS

try:
    from . import ptk
except ImportError:
    ptk = None

q('.py.pcc:`s#0 5 20f!32 33 31')
_prompt_color = q('{.py.pcc 100*(%)over system["w"]1 5}')
_prompt_namespace = q('{?[ns~`.;`;ns:system"d"]}')
if Q_OS.startswith('w'):
    def _colorize(_, prompt):
        return prompt
else:
    def _colorize(code, prompt):
        return "\001\033[%d;1m\002%s\001\033[0m\002" % (code, prompt)


class Cmd(_cmd.Cmd, object):
    """q REPL"""
    _prompt = 'q{ns})'

    @property
    def prompt(self):
        code = _prompt_color()
        prompt = self._prompt.format(ns=_prompt_namespace())
        return _colorize(code, prompt)

    def precmd(self, line):
        """Support for help"""
        if line.startswith('help'):
            if not q("`help in key`.q"):
                try:
                    q("\\l help.q")
                except kerr:
                    return '-1"no help available - install help.q"'
        if line == 'help':
            line += "`"
        return line

    def onecmd(self, line):
        """Interpret the line"""
        if line == '\\':
            return True
        elif line == 'EOF':
            print('\r', end='')
            return True
        else:
            try:
                v = q(line)
            except kerr as e:
                print("'%s" % e.args[0])
            else:
                if v != q('::'):
                    v.show()
            return False

    if ptk:
        cmdloop = ptk.cmdloop



from __future__ import absolute_import
from __future__ import unicode_literals

import ctypes
import subprocess
import sys
import platform
import shutil
import os
import pytest
import py

pytestmark = pytest.mark.skipif(platform.system() == "Windows",
                                reason="pyq.exe is not implemented")

linux_only = pytest.mark.skipif('linux' not in sys.platform.lower(),
                                reason="requires linux")


@pytest.fixture(scope='session')
def pyq_cmd():
    try:
        subprocess.check_call(['valgrind', '--version'])
    except (OSError, subprocess.CalledProcessError):
        return ['pyq']
    else:
        return ['valgrind', '-q', 'pyq']


def test_pyq_executable_success(pyq_cmd):
    out = subprocess.check_output(pyq_cmd + ['-V'],
                                  stderr=subprocess.STDOUT)
    version = [x for x in out.splitlines() if not x.startswith(b'profiling')]
    assert version[0].startswith(b'Python')

    executable = subprocess.check_output(
        ['pyq', '-c', 'import sys; print(sys.executable)'])
    assert executable.strip().endswith(b'pyq')


def test_pyq_executable_error(pyq_cmd):
    rc = subprocess.call(pyq_cmd + ['-c', 'raise SystemExit(42)'])
    assert rc == 42


def test_pyq_executable_versions(pyq_cmd):
    stream = 'stderr' if str is bytes else 'stdout'
    p = subprocess.Popen(pyq_cmd + ['--versions'],
                         **{stream: subprocess.PIPE})
    versions = getattr(p, stream).read()
    assert b'PyQ ' in versions
    assert b'KDB+ ' in versions
    assert b'Python ' in versions


@linux_only
@pytest.mark.parametrize('c, r', [
    ('0', 'n = 1\n0 \n'),
    ('1', 'n = 1\n1 \n'),
    ('1-2', 'n = 2\n1 2 \n'),
    ('1-3', 'n = 3\n1 2 3 \n'),
    ('1-4', 'n = 4\n1 2 3 4 \n'),
    ('1-5', 'n = 5\n1 2 3 4 5 \n'),
    ('1-6', 'n = 6\n1 2 3 4 5 6 \n'),
    ('1-7', 'n = 7\n1 2 3 4 5 6 7 \n'),
    ('1-8', 'n = 8\n1 2 3 4 5 6 7 8 \n'),
    ('1,3,5,7', 'n = 4\n1 3 5 7 \n'),
    ('2-4,8', 'n = 4\n2 3 4 8 \n'),
    ('0,4-8', 'n = 6\n0 4 5 6 7 8 \n'),
    ('0-3,4,7-8', 'n = 7\n0 1 2 3 4 7 8 \n'),
    ('0-3,4,7,10', 'n = 7\n0 1 2 3 4 7 10 \n'),
    ('x', 'n = 1\n0 \n'),
    ('0, 1, 2', 'n = 3\n0 1 2 \n'),
    ('0 , 1', 'n = 2\n0 1 \n'),
    ('0 ,1', 'n = 2\n0 1 \n'),
])
def test_pyq_taskset(c, r, monkeypatch, pyq_cmd):
    monkeypatch.setenv('CPUS', c)
    monkeypatch.setenv('TEST_CPUS', 'y')
    assert r == subprocess.check_output(pyq_cmd).decode()


def test_q_not_found(tmpdir, monkeypatch, pyq_cmd):
    monkeypatch.setenv('QHOME', str(tmpdir))
    monkeypatch.setenv('HOME', str(tmpdir.join('empty').ensure(dir=True)))
    monkeypatch.setenv("VIRTUAL_ENV", str(tmpdir))
    monkeypatch.setenv('PATH', str(tmpdir), prepend=':')
    tmpdir.join('pyq').mksymlinkto(sys.executable)
    p = subprocess.Popen(pyq_cmd, stderr=subprocess.PIPE)
    errors = p.stderr.read()
    rc = p.wait()
    assert rc != 0
    assert b"No such file or directory" in errors


@pytest.fixture
def q_arch(q):
    return str(q('.z.o'))


def test_q_venv0(tmpdir, monkeypatch, q_arch, pyq_cmd):
    # QHOME not set, $VIRTUAL_ENV/q present with q executable
    monkeypatch.setenv("VIRTUAL_ENV", str(tmpdir))
    monkeypatch.delenv("QHOME")
    venv = tmpdir.join('q')
    venv.join('python.q').ensure()
    q_exe = venv.join(q_arch, 'q')
    q_exe.write("#!/usr/bin/env bash\necho 'pass'", ensure=True)
    q_exe.chmod(0o755)
    p = subprocess.Popen(pyq_cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    err = p.stderr.readlines()
    assert err == []
    out = p.stdout.readlines()
    assert out[0].strip() == b'pass'


def test_q_venv1(tmpdir, monkeypatch, pyq_cmd):
    # QHOME not set, $VIRTUAL_ENV/q present, but no q executable
    monkeypatch.setenv("VIRTUAL_ENV", str(tmpdir))
    monkeypatch.setenv("HOME", str(tmpdir.join("empty").ensure(dir=True)))
    monkeypatch.delenv("QHOME")
    monkeypatch.setenv('PATH', str(tmpdir), prepend=':')
    py.path.local(sys.executable).copy(tmpdir.join('pyq'), mode=True)
    qhome = tmpdir.join('q').ensure(dir=True)
    p = subprocess.Popen(pyq_cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    err = p.stderr.read()
    out = p.stdout.read()
    rc = p.wait()
    assert out == b''
    assert qhome.dirname.encode() in err
    assert rc != 0


def test_q_venv2(tmpdir, monkeypatch, q_arch, pyq_cmd):
    # QHOME not set, VIRTUAL_ENV not set, HOME set to tempdir
    monkeypatch.setenv("HOME", str(tmpdir))
    monkeypatch.delenv("QHOME")
    monkeypatch.delenv("VIRTUAL_ENV", raising=False)
    monkeypatch.setenv('PATH', str(tmpdir), prepend=':')
    py.path.local(sys.executable).copy(tmpdir.join('pyq'), mode=True)

    q_exe = tmpdir.join('q', q_arch, 'q')
    q_exe.write("#!/usr/bin/env bash\necho 'pass'", ensure=True)
    q_exe.chmod(0o755)

    p = subprocess.Popen(pyq_cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out = p.stdout.readlines()
    err = p.stderr.readlines()
    assert out[0].strip() == b'pass'
    assert err == []
    assert p.wait() == 0


def test_q_venv3(tmpdir, monkeypatch, q_arch, pyq_cmd):
    # QHOME not set, VIRTUAL_ENV not set, q is next to bin/pyq
    monkeypatch.delenv("QHOME")
    monkeypatch.delenv("VIRTUAL_ENV", raising=False)
    bindir = tmpdir.join('bin')
    bindir.ensure(dir=1)
    monkeypatch.setenv('PATH', str(bindir), prepend=':')
    py.path.local(sys.executable).copy(bindir.join('pyq'), mode=True)

    q_exe = tmpdir.join('q', q_arch, 'q')
    q_exe.write("#!/usr/bin/env bash\necho 'pass'", ensure=True)
    q_exe.chmod(0o755)

    p = subprocess.Popen(pyq_cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out = p.stdout.readlines()
    err = p.stderr.readlines()
    assert out[0].strip() == b'pass'
    assert err == []
    assert p.wait() == 0


def test_pyq_trace(pyq_cmd):
    output = subprocess.check_output(pyq_cmd + ['--pyq-trace', '-c', '0'])
    assert b"pyq trace is on" in output


def test_pyq_preload(tmpdir, q, pyq_cmd):
    db = tmpdir.join('db')
    q.x = q.til(3)
    q.save(db.join('x'))
    output = subprocess.check_output(pyq_cmd + [
        str(db), "-c", "from pyq import q; print(q.x)"])
    assert output == b"0 1 2\n"


def test_p__file__0(tmpdir, pyq_cmd):
    p = tmpdir.join('test.py')
    p.write("print(__file__)")
    out = subprocess.check_output(pyq_cmd + [str(p)])
    assert out.strip().endswith(str(p).encode())


def test_broken_q(tmpdir, monkeypatch, q_arch, pyq_cmd):
    # QHOME not set, $VIRTUAL_ENV/q present with a broken q executable
    monkeypatch.setenv("VIRTUAL_ENV", str(tmpdir))
    monkeypatch.delenv("QHOME")
    q_exe = tmpdir.join('q', q_arch, 'q')
    q_exe.ensure()
    q_exe.chmod(0o422)
    with open(os.devnull, 'w')as null:
        p = subprocess.Popen(pyq_cmd, stdout=null, stderr=subprocess.PIPE)
        assert (str(q_exe) + ':') in p.stderr.read().decode()


@pytest.mark.parametrize('message,stock', [
    ('standalone python', True),
    ('', False),
]
                         )
def test_stock_python(monkeypatch, message, stock):
    if stock:
        monkeypatch.setattr(ctypes, 'CDLL', lambda x: None)
    monkeypatch.setitem(sys.modules, 'pyq._k', None)
    monkeypatch.delitem(sys.modules, 'pyq', raising=False)

    with pytest.raises(ImportError) as exc:
        __import__('pyq')

    if stock:
        assert message in exc.value.args[0]


def test_dash_dash(pyq_cmd):
    # NB: -P is q command line option to set display precision
    output = subprocess.check_output(pyq_cmd + ['-c', r'print(q(r"\P"))',
                                                '--', '-P', '10'])
    assert b'10' in output


@pytest.mark.parametrize('option, message', [
    ('qhome', b'QHOME='),
    ('qarch', b'QARCH='),
    ('qbin', b'QBIN='),
])
def test_print_options(option, message):
    output = subprocess.check_output(['pyq', '--pyq-' + option])
    assert message in output
    value = 'xyz'
    output = subprocess.check_output(['pyq', '--pyq-' + option,
                                      '--pyq-' + option + '=' + value])
    assert (message + value.encode()) in output


@pytest.mark.skipif("not hasattr(shutil, 'which')")
def test_run_with_absolute_path():
    abs_path = shutil.which('pyq')
    out = subprocess.check_output([abs_path, '-c',
                                   'import sys;print(sys.executable)'])
    assert os.stat(out.strip().decode()) == os.stat(abs_path)


@pytest.mark.skipif("not hasattr(shutil, 'which')")
def test_run_with_relative_path(monkeypatch):
    abs_path = shutil.which('pyq')
    monkeypatch.chdir(os.path.dirname(abs_path))
    out = subprocess.check_output(['./pyq', '-c',
                                   'import sys;print(sys.executable)'])
    assert os.stat(out.strip().decode()) == os.stat(abs_path)


@pytest.mark.skipif("not hasattr(shutil, 'which')")
def test_run_with_single_dir_in_path(monkeypatch):
    abs_path = shutil.which('pyq')
    monkeypatch.setenv('PATH', os.path.dirname(abs_path))
    out = subprocess.check_output(['pyq', '-c',
                                   'import sys;print(sys.executable)'])
    assert os.stat(out.strip().decode()) == os.stat(abs_path)



"""PyQ - Python for kdb+

PyQ brings the Python programming language to the kdb+ database.  It allows
developers to seamlessly integrate Python and q codes in one application.
This is achieved by bringing the Python and q interpreters in the same
process so that codes written in either of the languages operate on the same
data.  In PyQ, Python and q objects live in the same memory space and share
the same data.

"""
import os
import platform
import subprocess
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from distutils.core import Command, Distribution, Extension
from distutils.command.build import build
from distutils.command.build_ext import build_ext
from distutils.command.config import config
from distutils.command.install import install
from distutils.command.install_data import install_data
from distutils.command.install_scripts import install_scripts

import sysconfig

WINDOWS = platform.system() == 'Windows'
VERSION_FILE = 'src/pyq/version.py'
CFLAGS = ['/WX', '/wd4090'] if WINDOWS else ['-fno-strict-aliasing']

if sys.version_info >= (3, ) and not WINDOWS:
    CFLAGS.append('-Werror')

LDFLAGS = []
if (sys.maxsize + 1).bit_length() == 32 and platform.machine() == 'x86_64':
    # Building 32-bit pyq on a 64-bit host
    config_vars = sysconfig.get_config_vars()
    CFLAGS.append('-m32')
    LDFLAGS.append('-m32')

    def split_replace(string, a, b, sep):
        x = string.split(sep)
        for i, part in enumerate(x):
            if part == a:
                x[i] = b
        return sep.join(x)

    for k, v in config_vars.items():
        if isinstance(v, str):
            config_vars[k] = split_replace(v, 'x86_64', 'i386', '-')

TEST_REQUIREMENTS = [
                        'pytest>=2.6.4,!=3.2.0,!=3.3.0',
                        'pytest-pyq',
                        'pytest-cov>=2.4',
                        'coverage>=4.2'
                    ] + (['pathlib2>=2.0'] if sys.version_info[0] < 3 else [])

IPYTHON_REQUIREMENTS = ['ipython']

Executable = Extension

METADATA = dict(
    name='pyq',
    packages=['pyq', 'pyq.tests', ],
    package_dir={'': 'src'},
    qlib_scripts=['python.q', 'p.k', 'pyq-operators.q', 'pyq-print.q', ],
    ext_modules=[
        Extension('pyq._k', sources=['src/pyq/_k.c', ],
                  extra_compile_args=CFLAGS,
                  extra_link_args=LDFLAGS),
    ],
    qext_modules=[
        Extension('pyq', sources=['src/pyq/pyq.c', ],
                  extra_compile_args=CFLAGS,
                  extra_link_args=LDFLAGS),
    ],
    executables=[] if WINDOWS else [
        Executable('pyq', sources=['src/pyq.c'],
                   extra_compile_args=CFLAGS,
                   extra_link_args=LDFLAGS),
    ],
    scripts=['src/scripts/pyq-runtests',
             'src/scripts/pyq-coverage',
             'src/scripts/ipyq',
             'src/scripts/pq',
             'src/scripts/qp',
             ],
    data_files=[
        ('q', ['src/pyq/p.k',
               'src/pyq/pyq-operators.q',
               'src/pyq/python.q',
               ]
         ),
    ],
    url='https://github.com/KxSystems/pyq',
    maintainer='PyQ Authors',
    maintainer_email='pyq@enlnt.com',
    license='Apache License',
    platforms=['Linux', 'MacOS X', 'Windows'],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Financial and Insurance Industry',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: Apache Software License',
                 'Natural Language :: English',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Microsoft :: Windows :: Windows 10',
                 'Programming Language :: C',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Topic :: Database',
                 'Topic :: Software Development :: Libraries' +
                 ' :: Python Modules'],
)


def add_data_file(data_files, target, source):
    """Add an entry to data_files"""
    for t, f in data_files:
        if t == target:
            break
    else:
        data_files.append((target, []))
        f = data_files[-1][1]
    if source not in f:
        f.append(source)


def get_q_home(env):
    """Derive q home from the environment"""
    q_home = env.get('QHOME')
    if q_home:
        return q_home
    for v in ['VIRTUAL_ENV', 'HOME']:
        prefix = env.get(v)
        if prefix:
            q_home = os.path.join(prefix, 'q')
            if os.path.isdir(q_home):
                return q_home
    if WINDOWS:
        q_home = os.path.join(env['SystemDrive'], r'\q')
        if os.path.isdir(q_home):
            return q_home
    raise RuntimeError('No suitable QHOME.')


def get_q_os_letter(sysname, machine):
    if sysname == 'Linux':
        return 'l'
    if sysname == 'SunOS':
        return 'v' if machine == 'i86pc' else 's'
    if sysname == 'Darwin':
        return 'm'
    if sysname == 'Windows':
        return 'w'
    raise RuntimeError('"Unknown platform: %s %s.' % (sysname, machine))


def get_q_arch(q_home):
    bits = (sys.maxsize + 1).bit_length()
    sysname = platform.system()
    machine = platform.machine()

    os_letter = get_q_os_letter(sysname, machine)
    if bits == 64:
        # In case we're on 64-bit platform, but 64-bit kdb+ is not available
        # we will fallback to the 32-bit version.
        x64dir = os.path.join(q_home, '%s64' % os_letter)
        if not os.path.isdir(x64dir):
            bits = 32
    return '%s%d' % (os_letter, bits)


def get_q_version(q_home):
    """Return version of q installed at q_home"""
    with open(os.path.join(q_home, 'q.k')) as f:
        for line in f:
            if line.startswith('k:'):
                return line[2:5]
    return '2.2'


decode = (lambda x: x) if str is bytes else lambda x: x.decode()


def get_python_dll(executable):
    sysname = platform.system()
    if sysname.startswith(('Linux', 'SunOS')):
        output = subprocess.check_output(['ldd', executable])
        for line in output.splitlines():
            if b'libpython' in line:
                return decode(line.split()[2])

        # This is for systems which have statically linked Python
        # (i.e Ubuntu), but provide dynamic libraries in a separate
        # package.
        libpython = 'libpython{}.{}'.format(*sys.version_info[:2]).encode()
        try:
            output = subprocess.check_output(['ldconfig', '-p'])
        except subprocess.CalledProcessError:
            output = subprocess.check_output(['/sbin/ldconfig', '-p'])
        for line in output.splitlines():
            if libpython in line:
                return decode(line.split()[-1])

    elif sysname == 'Darwin':
        output = subprocess.check_output(['otool', '-L', executable])
        for line in output.splitlines()[1:]:
            if b'Python' in line:
                python_dll = decode(line.split()[0])
                return python_dll.replace('@executable_path',
                                          os.path.dirname(executable))
    elif sysname == 'Windows':
        return 'python{}{}.dll'.format(*sys.version_info[:2])
    # This is known to work for Anaconda
    ldlibrary = sysconfig.get_config_var('LDLIBRARY')
    libdir = sysconfig.get_config_var('LIBDIR')
    if ldlibrary and libdir:
        libfile = os.path.join(libdir, ldlibrary)
        if os.path.exists(libfile):
            return libfile
    raise RuntimeError('no python dll')


SETUP_CFG = """\
[config]
q_home = {q_home}
q_version = {q_version}
q_arch = {q_arch}
python_dll = {python_dll}
"""


class Config(config):
    user_options = [
        ('q-home=', None, 'q home directory'),
        ('q-version=', None, 'q version'),
        ('q-arch=', None, 'q architecture, e.g. l64'),
        ('python-dll=', None, 'path to the python dynamic library'),
        ('dest=', None, "path to the config file (default: setup.cfg)"),
        ('write', None, 'write the config file')
    ]

    q_home = None
    q_arch = None
    q_version = None
    python_dll = None
    dest = None
    write = None

    extra_link_args = []

    def initialize_options(self):
        config.initialize_options(self)

    def finalize_options(self):
        if self.q_home is None:
            self.q_home = get_q_home(os.environ)
        if self.q_arch is None:
            self.q_arch = get_q_arch(self.q_home)
        if self.q_version is None:
            self.q_version = get_q_version(self.q_home)
        if self.python_dll is None:
            self.python_dll = get_python_dll(sys.executable)
        if self.dest is None:
            self.dest = 'setup.cfg'
        if WINDOWS:
            self.extra_link_args = [r'src\pyq\kx\%s\q.lib' % self.q_arch]

    def run(self):
        setup_cfg = SETUP_CFG.format(**vars(self))
        self.announce(setup_cfg.rstrip(), 2)
        if self.write:
            with open(self.dest, 'w') as f:
                f.write(setup_cfg)
            self.announce('^^^ Written to %s.' % self.dest, 2)
        else:
            self.announce('^^^ Use --write options'
                          ' to write this to %s.' % self.dest, 2)


PYQ_CONFIG = """\
\\d .pyq
python_dll:"{python_dll}\\000"
pyq_executable:"{pyq_executable}"
"""


class BuildQLib(Command):
    description = "build q/k scripts"

    user_options = [
        ('build-lib=', 'd', "build directory"),
        ('force', 'f', "forcibly build everything (ignore file timestamps)"),
    ]

    q_home = None
    build_base = None
    build_lib = None

    python_dll = None
    pyq_executable = None

    def initialize_options(self):
        pass

    def finalize_options(self):
        self.set_undefined_options('config',
                                   ('q_home', 'q_home'),
                                   ('python_dll', 'python_dll'))
        self.set_undefined_options('build',
                                   ('build_base', 'build_base'))
        self.build_lib = os.path.join(self.build_base, 'qlib')
        cmd = self.get_finalized_command('install_exe')
        pyq_path = os.path.join(cmd.install_dir, 'pyq')
        self.pyq_executable = pyq_path.replace('\\', '\\\\')

    def run(self):
        self.mkpath(self.build_lib)
        for script in self.distribution.qlib_scripts:
            outfile = os.path.join(self.build_lib, script)
            script_file = os.path.join('src', 'pyq', script)
            self.write_pyq_config()
            self.copy_file(script_file, outfile, preserve_mode=0)

    def write_pyq_config(self):
        pyq_config_file = os.path.join(self.build_lib, 'pyq-config.q')
        with open(pyq_config_file, 'w') as f:
            f.write(PYQ_CONFIG.format(**vars(self)))
        add_data_file(self.distribution.data_files, 'q', pyq_config_file)


class BuildQExt(Command):
    description = "build q extension modules"

    user_options = [
        ('build-lib=', 'd', "build directory"),
        ('force', 'f', "forcibly build everything (ignore file timestamps)"),
    ]

    q_home = None
    q_arch = None
    q_version = None

    build_base = None
    build_temp = None
    build_lib = None

    compiler = None
    define = None
    debug = None
    force = None
    plat_name = None

    extensions = None

    def initialize_options(self):
        pass

    def finalize_options(self):
        self.set_undefined_options('config',
                                   ('q_home', 'q_home'),
                                   ('q_arch', 'q_arch'),
                                   ('q_version', 'q_version'))
        self.set_undefined_options('build',
                                   ('build_base', 'build_base'),
                                   ('compiler', 'compiler'),
                                   ('debug', 'debug'),
                                   ('force', 'force'),
                                   ('plat_name', 'plat_name'))
        if self.build_lib is None:
            self.build_lib = os.path.join(self.build_base,
                                          'qext.' + self.plat_name)
        if self.build_temp is None:
            self.build_temp = os.path.join(self.build_base,
                                           'temp.' + self.plat_name)
        if self.extensions is None:
            self.extensions = self.distribution.qext_modules
        if self.define is None:
            split_version = self.q_version.split('.')
            self.define = [('KXVER', split_version[0]),
                           ('KXVER2', split_version[1]), ]

    def run(self):
        from distutils.ccompiler import new_compiler
        from distutils.sysconfig import customize_compiler

        include_dirs = ['src/pyq/kx', ]

        conf = self.get_finalized_command("config")
        for ext in self.extensions:
            sources = ext.sources
            ext_path = os.path.join(self.build_lib,
                                    ext.name + ('.dll' if WINDOWS else '.so'))
            compiler = new_compiler(compiler=self.compiler,
                                    verbose=self.verbose,
                                    dry_run=self.dry_run,
                                    force=self.force)
            customize_compiler(compiler)
            if WINDOWS:
                compiler.initialize()
                compiler.compile_options.remove('/MD')
            extra_args = ext.extra_compile_args or []
            objects = compiler.compile(sources,
                                       output_dir=self.build_temp,
                                       macros=self.define,
                                       extra_postargs=extra_args,
                                       include_dirs=include_dirs)
            extra_args = conf.extra_link_args[:] + ext.extra_link_args
            if WINDOWS:
                extra_args.extend([r'/DEF:src\pyq\%s.def' % ext.name])

            compiler.link_shared_object(objects, ext_path,
                                        extra_postargs=extra_args)
            add_data_file(self.distribution.data_files,
                          os.path.join('q', self.q_arch), ext_path)


class BuildPyExt(build_ext):
    q_arch = None

    def finalize_options(self):
        build_ext.finalize_options(self)
        self.set_undefined_options('build_qext',
                                   ('define', 'define'))
        self.set_undefined_options('config',
                                   ('q_arch', 'q_arch'))
        conf = self.get_finalized_command("config")
        if conf.extra_link_args:
            for ext in self.extensions:
                ext.extra_link_args = [a.format(**vars(ext))
                                       for a in conf.extra_link_args]

    if WINDOWS:
        def build_extensions(self):
            self.compiler.initialize()
            self.compiler.compile_options.remove('/MD')
            build_ext.build_extensions(self)


class BuildExe(Command):
    description = "build executables"
    user_options = []

    q_home = None
    q_arch = None
    q_version = None

    build_temp = None
    build_exe = None
    build_base = None

    compiler = None
    debug = None
    define = None
    plat_name = None

    def initialize_options(self):
        pass

    def finalize_options(self):
        self.set_undefined_options('config',
                                   ('q_home', 'q_home'),
                                   ('q_arch', 'q_arch'),
                                   ('q_version', 'q_version'))
        self.set_undefined_options('build',
                                   ('build_base', 'build_base'),
                                   ('compiler', 'compiler'),
                                   ('debug', 'debug'),
                                   ('force', 'force'),
                                   ('plat_name', 'plat_name'))

        if self.build_exe is None:
            self.build_exe = os.path.join(self.build_base,
                                          'exe.{}-{}'.format(self.plat_name,
                                                             sys.version[:3]))
        if self.define is None:
            self.define = [
                ('KXVER', self.q_version[0]),
                ('QARCH', self.q_arch),
            ]

    def run(self):
        from distutils.ccompiler import new_compiler
        from distutils.sysconfig import customize_compiler

        for exe in self.distribution.executables:
            compiler = new_compiler(
                compiler=self.compiler,
                verbose=self.verbose,
                dry_run=self.dry_run,
                force=self.force)
            customize_compiler(compiler)
            extra_args = exe.extra_compile_args or []
            objects = compiler.compile(exe.sources,
                                       macros=self.define,
                                       extra_postargs=extra_args,
                                       output_dir=self.build_temp)
            compiler.link_executable(objects,
                                     extra_preargs=LDFLAGS,
                                     output_progname=exe.name,
                                     output_dir=self.build_exe)


class InstallQLib(install_data):
    description = "install q/k scripts"

    build_dir = None
    skip_build = None
    outfiles = None

    def finalize_options(self):
        self.set_undefined_options('config', ('q_home', 'install_dir'))
        self.set_undefined_options('build_qlib', ('build_lib', 'build_dir'))
        self.set_undefined_options('install', ('skip_build', 'skip_build'))

    def run(self):
        if not self.skip_build:
            self.run_command('build_qlib')
        self.outfiles = self.copy_tree(self.build_dir, self.install_dir)


class InstallQExt(install_data):
    description = "install q/k scripts"

    q_home = None
    q_arch = None
    build_dir = None
    skip_build = None
    install_dir = None
    outfiles = None

    def finalize_options(self):
        self.set_undefined_options('config',
                                   ('q_home', 'q_home'),
                                   ('q_arch', 'q_arch'))
        self.set_undefined_options('build_qext', ('build_lib', 'build_dir'))
        self.set_undefined_options('install', ('skip_build', 'skip_build'))
        self.install_dir = os.path.join(self.q_home, self.q_arch)

    def run(self):
        if not self.skip_build:
            self.run_command('build_qext')
        self.outfiles = self.copy_tree(self.build_dir, self.install_dir)


class InstallExe(install_scripts):
    description = "install executables"

    outfiles = None

    def finalize_options(self):
        self.set_undefined_options('build_exe', ('build_exe', 'build_dir'))
        self.set_undefined_options('install',
                                   ('install_scripts', 'install_dir'),
                                   ('force', 'force'),
                                   ('skip_build', 'skip_build'),
                                   )

    def run(self):
        if not self.skip_build:
            self.run_command('build_exe')
        from stat import ST_MODE
        if not self.get_inputs():
            return
        self.outfiles = self.copy_tree(self.build_dir, self.install_dir)
        if os.name == 'posix':
            # Set the executable bits (owner, group, and world) on
            # all the executables we just installed.
            for file in self.get_outputs():
                if self.dry_run:
                    self.announce("changing mode of %s" % file, 2)
                else:
                    mode = ((os.stat(file)[ST_MODE]) | 0o555) & 0o7777
                    self.announce("changing mode of %s to %o" %
                                  (file, mode), 2)
                    os.chmod(file, mode)

    def get_inputs(self):
        return self.distribution.executables or []


class PyqDistribution(Distribution):
    qlib_scripts = None
    qext_modules = None
    executables = None


build.sub_commands.extend([
    ('build_qlib', None),
    ('build_qext', None),
    ('build_exe', None),
])

install.sub_commands.extend([
    ('install_qlib', None),
    ('install_qext', None),
    ('install_exe', None),
])


def run_setup(metadata):
    summary, details = __doc__.split('\n\n', 1)
    with open('README.md') as f:
        long_description = f.read()
    keywords = metadata.copy()
    keywords.update(
        use_scm_version={'write_to': VERSION_FILE},
        description=summary,
        long_description=long_description,
        long_description_content_type='text/markdown',
        distclass=PyqDistribution,
        cmdclass={
            'config': Config,
            'build_qlib': BuildQLib,
            'build_qext': BuildQExt,
            'build_ext': BuildPyExt,
            'build_exe': BuildExe,
            'install_qlib': InstallQLib,
            'install_qext': InstallQExt,
            'install_exe': InstallExe,
        },
    )
    if 'setuptools' in sys.modules:
        keywords['extras_require'] = {
            'test': TEST_REQUIREMENTS,
            'ipython': IPYTHON_REQUIREMENTS,
            'all': TEST_REQUIREMENTS + IPYTHON_REQUIREMENTS + [
                'py', 'numpy', 'prompt-toolkit', 'pygments-q'],
        }
        keywords['setup_requires'] = ['setuptools_scm']

    setup(**keywords)


if __name__ == '__main__':
    run_setup(METADATA)



# These tests can only be run from the package source directory
import pytest
import os
import sys
import subprocess
import re

pytestmark = pytest.mark.skipif("not os.path.exists('setup.py')")

python = sys.executable

version_re = re.compile(r'^\d\.\d.*$')  # TODO: Come up with a more restrictive RE.

def test_version():
    v = subprocess.check_output([python, 'setup.py', '--version']).decode()
    assert version_re.match(v)

# For now, just test that the build succeeds.
# TODO: test under different scenarios w.r.t. QHOME, etc.
def test_build():
    rc = subprocess.call([python, 'setup.py', 'build'])
    assert rc == 0

def test_config(tmpdir):
    dest = tmpdir.join('setup.cfg')
    rc = subprocess.call([python, 'setup.py', 'config',
                          '--python-dll', 'faked',
                          '--dest', dest.strpath,
                          '--write'])
    assert rc == 0
    assert dest.read().startswith('[config]')



from __future__ import absolute_import
import sys
import pytest


@pytest.fixture
def no_numpy_pyq(monkeypatch):
    monkeypatch.setitem(sys.modules, 'numpy', None)
    pyq_modules = [m for m in sys.modules
                   if m is not None and (m == 'pyq' or m.startswith('pyq.'))]
    for m in pyq_modules:
        monkeypatch.delitem(sys.modules, m)
    import pyq
    return pyq


def test_no_numpy(no_numpy_pyq):
    # TODO: Figure out how to run all tests that don't require numpy.
    assert no_numpy_pyq.K(0) == 0



from __future__ import absolute_import
import subprocess
import os
import pytest
from pyq import _k
import platform

WIN = platform.system() == "Windows"

TEST_SD = """\
p)import os
p)import sys
p)from pyq import _k
p)def f(d):
    x = os.read(d, 1)
    print(x.decode())
    sys.stdout.flush()
    _k.sd0(d)
p)r,w = os.pipe()
p)_k.sd1(r, f)
p)os.write(w, b'X')
"""

TEST_SD_ERROR = """\
p)from pyq import _k
p)r,w = os.pipe()
p)_k.sd1(r, None)
p)os.write(w, b'X')
"""


@pytest.mark.skipif(WIN, reason="This test hangs on Windows.")
def test_sd(tmpdir):
    qbin = os.environ['QBIN']
    test = tmpdir.join('test_sd.q')
    test.write(TEST_SD)
    with open(os.devnull) as devnull:
        assert b'X\n' == subprocess.check_output([qbin, str(test)],
                                                 stdin=devnull)


@pytest.mark.skipif(WIN, reason="This test hangs on Windows.")
def test_callback_error(tmpdir):
    qbin = os.environ['QBIN']
    test = tmpdir.join('test_sd.q')
    test.write(TEST_SD_ERROR)
    with open(os.devnull) as devnull:
        p = subprocess.Popen([qbin, str(test)],
                             stdin=devnull, stderr=subprocess.PIPE)
        out, err = p.communicate()
    assert b'py' in err


def test_sd1_error():
    with pytest.raises(ValueError):
        _k.sd1(-1, None)
